/* Generated by Cython 0.25.2 */

/* BEGIN: Cython Metadata
{
    "distutils": {},
    "module_name": "neo4j.packstream._packer"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_25_2"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x03030000 || (PY_MAJOR_VERSION == 2 && PY_VERSION_HEX >= 0x02070000)
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_FASTCALL
  #define METH_FASTCALL 0x80
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject **args,
                                              Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__neo4j__packstream___packer
#define __PYX_HAVE_API__neo4j__packstream___packer
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "neo4j/packstream/_packer.pyx",
};

/*--- Type declarations ---*/
struct __pyx_obj_5neo4j_10packstream_7_packer_Packer;

/* "neo4j/packstream/_packer.pyx":34
 * 
 * 
 * cdef class Packer(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef public bint supports_bytes
 */
struct __pyx_obj_5neo4j_10packstream_7_packer_Packer {
  PyObject_HEAD
  struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *__pyx_vtab;
  int supports_bytes;
  PyObject *stream;
  PyObject *_write;
};



struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer {
  PyObject *(*pack_raw)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, PyObject *);
  PyObject *(*pack)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_pack)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, PyObject *);
  PyObject *(*pack_bytes_header)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int);
  PyObject *(*pack_string_header)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int);
  PyObject *(*pack_list_header)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int);
  PyObject *(*pack_list_stream_header)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int __pyx_skip_dispatch);
  PyObject *(*pack_map_header)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int);
  PyObject *(*pack_map_stream_header)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int __pyx_skip_dispatch);
  PyObject *(*pack_struct)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*pack_end_of_stream)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *__pyx_vtabptr_5neo4j_10packstream_7_packer_Packer;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_RemainderObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_RemainderObjC(op1, op2, intval, inplace)\
    (inplace ? PyNumber_InPlaceRemainder(op1, op2) : PyNumber_Remainder(op1, op2))
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ArgTypeTest.proto */
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CStringEquals.proto */
static CYTHON_INLINE int __Pyx_StrEq(const char *, const char *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_raw(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_data); /* proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_value, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer__pack(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_bytes_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_v_size); /* proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_string_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_v_size); /* proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_list_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_v_size); /* proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_list_stream_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_map_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_v_size); /* proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_map_stream_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_struct(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_signature, PyObject *__pyx_v_fields, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_end_of_stream(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'neo4j.packstream._packer' */
static PyTypeObject *__pyx_ptype_5neo4j_10packstream_7_packer_Packer = 0;
static PyObject *__pyx_v_5neo4j_10packstream_7_packer_INT64_HI = 0;
static PyObject *__pyx_v_5neo4j_10packstream_7_packer_INT64_LO = 0;
#define __Pyx_MODULE_NAME "neo4j.packstream._packer"
int __pyx_module_is_main_neo4j__packstream___packer = 0;

/* Implementation of 'neo4j.packstream._packer' */
static PyObject *__pyx_builtin_OverflowError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_TypeError;
static const char __pyx_k_[] = "\300";
static const char __pyx_k_I[] = ">I";
static const char __pyx_k_d[] = ">d";
static const char __pyx_k_i[] = ">i";
static const char __pyx_k_q[] = ">q";
static const char __pyx_k__3[] = "\303";
static const char __pyx_k__5[] = "\302";
static const char __pyx_k__7[] = "\301";
static const char __pyx_k__9[] = "\310";
static const char __pyx_k__11[] = "\311";
static const char __pyx_k__13[] = "\312";
static const char __pyx_k__15[] = "\313";
static const char __pyx_k__20[] = "\314";
static const char __pyx_k__22[] = "\315";
static const char __pyx_k__24[] = "\316";
static const char __pyx_k__27[] = "\200";
static const char __pyx_k__29[] = "\201";
static const char __pyx_k__31[] = "\202";
static const char __pyx_k__33[] = "\203";
static const char __pyx_k__35[] = "\204";
static const char __pyx_k__37[] = "\205";
static const char __pyx_k__39[] = "\206";
static const char __pyx_k__41[] = "\207";
static const char __pyx_k__43[] = "\210";
static const char __pyx_k__45[] = "\211";
static const char __pyx_k__47[] = "\212";
static const char __pyx_k__49[] = "\213";
static const char __pyx_k__51[] = "\214";
static const char __pyx_k__53[] = "\215";
static const char __pyx_k__55[] = "\216";
static const char __pyx_k__57[] = "\217";
static const char __pyx_k__59[] = "\320";
static const char __pyx_k__61[] = "\321";
static const char __pyx_k__63[] = "\322";
static const char __pyx_k__66[] = "\220";
static const char __pyx_k__68[] = "\221";
static const char __pyx_k__70[] = "\222";
static const char __pyx_k__72[] = "\223";
static const char __pyx_k__74[] = "\224";
static const char __pyx_k__76[] = "\225";
static const char __pyx_k__78[] = "\226";
static const char __pyx_k__80[] = "\227";
static const char __pyx_k__82[] = "\230";
static const char __pyx_k__84[] = "\231";
static const char __pyx_k__86[] = "\232";
static const char __pyx_k__88[] = "\233";
static const char __pyx_k__90[] = "\234";
static const char __pyx_k__92[] = "\235";
static const char __pyx_k__94[] = "\236";
static const char __pyx_k__96[] = "\237";
static const char __pyx_k__98[] = "\324";
static const char __pyx_k__100[] = "\325";
static const char __pyx_k__102[] = "\326";
static const char __pyx_k__105[] = "\327";
static const char __pyx_k__107[] = "\240";
static const char __pyx_k__109[] = "\241";
static const char __pyx_k__111[] = "\242";
static const char __pyx_k__113[] = "\243";
static const char __pyx_k__115[] = "\244";
static const char __pyx_k__117[] = "\245";
static const char __pyx_k__119[] = "\246";
static const char __pyx_k__121[] = "\247";
static const char __pyx_k__123[] = "\250";
static const char __pyx_k__125[] = "\251";
static const char __pyx_k__127[] = "\252";
static const char __pyx_k__129[] = "\253";
static const char __pyx_k__131[] = "\254";
static const char __pyx_k__133[] = "\255";
static const char __pyx_k__135[] = "\256";
static const char __pyx_k__137[] = "\257";
static const char __pyx_k__139[] = "\330";
static const char __pyx_k__141[] = "\331";
static const char __pyx_k__143[] = "\332";
static const char __pyx_k__146[] = "\333";
static const char __pyx_k__149[] = "\260";
static const char __pyx_k__151[] = "\261";
static const char __pyx_k__153[] = "\262";
static const char __pyx_k__155[] = "\263";
static const char __pyx_k__157[] = "\264";
static const char __pyx_k__159[] = "\265";
static const char __pyx_k__161[] = "\266";
static const char __pyx_k__163[] = "\267";
static const char __pyx_k__165[] = "\270";
static const char __pyx_k__167[] = "\271";
static const char __pyx_k__169[] = "\272";
static const char __pyx_k__171[] = "\273";
static const char __pyx_k__173[] = "\274";
static const char __pyx_k__175[] = "\275";
static const char __pyx_k__177[] = "\276";
static const char __pyx_k__179[] = "\277";
static const char __pyx_k__181[] = "\334";
static const char __pyx_k__183[] = "\335";
static const char __pyx_k__186[] = "\337";
static const char __pyx_k__188[] = "*";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_fields[] = "fields";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_stream[] = "stream";
static const char __pyx_k_string[] = "string";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_integer[] = "integer";
static const char __pyx_k_Structure[] = "Structure";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_constants[] = "constants";
static const char __pyx_k_signature[] = "signature";
static const char __pyx_k_structure[] = "structure";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_pack_struct[] = "pack_struct";
static const char __pyx_k_struct_pack[] = "struct_pack";
static const char __pyx_k_neo4j_compat[] = "neo4j.compat";
static const char __pyx_k_OverflowError[] = "OverflowError";
static const char __pyx_k_PACKED_UINT_8[] = "PACKED_UINT_8";
static const char __pyx_k_PACKED_UINT_16[] = "PACKED_UINT_16";
static const char __pyx_k_pack_end_of_stream[] = "pack_end_of_stream";
static const char __pyx_k_Integer_s_out_of_range[] = "Integer %s out of range";
static const char __pyx_k_pack_map_stream_header[] = "pack_map_stream_header";
static const char __pyx_k_pack_list_stream_header[] = "pack_list_stream_header";
static const char __pyx_k_Structure_size_out_of_range[] = "Structure size out of range";
static const char __pyx_k_Map_header_size_out_of_range[] = "Map header size out of range";
static const char __pyx_k_List_header_size_out_of_range[] = "List header size out of range";
static const char __pyx_k_Bytes_header_size_out_of_range[] = "Bytes header size out of range";
static const char __pyx_k_String_header_size_out_of_range[] = "String header size out of range";
static const char __pyx_k_Structures_require_a_2_tuple_of[] = "Structures require a 2-tuple of (signature, fields)";
static const char __pyx_k_This_PackSteam_channel_does_not[] = "This PackSteam channel does not support BYTES (consider upgrading to Neo4j 3.2+)";
static const char __pyx_k_Structure_signature_must_be_a_si[] = "Structure signature must be a single byte value";
static const char __pyx_k_Values_of_type_s_are_not_support[] = "Values of type %s are not supported";
static PyObject *__pyx_kp_b_;
static PyObject *__pyx_kp_s_Bytes_header_size_out_of_range;
static PyObject *__pyx_kp_s_I;
static PyObject *__pyx_kp_s_Integer_s_out_of_range;
static PyObject *__pyx_kp_s_List_header_size_out_of_range;
static PyObject *__pyx_kp_s_Map_header_size_out_of_range;
static PyObject *__pyx_n_s_OverflowError;
static PyObject *__pyx_n_s_PACKED_UINT_16;
static PyObject *__pyx_n_s_PACKED_UINT_8;
static PyObject *__pyx_kp_s_String_header_size_out_of_range;
static PyObject *__pyx_n_s_Structure;
static PyObject *__pyx_kp_s_Structure_signature_must_be_a_si;
static PyObject *__pyx_kp_s_Structure_size_out_of_range;
static PyObject *__pyx_kp_s_Structures_require_a_2_tuple_of;
static PyObject *__pyx_kp_s_This_PackSteam_channel_does_not;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s_Values_of_type_s_are_not_support;
static PyObject *__pyx_kp_b__100;
static PyObject *__pyx_kp_b__102;
static PyObject *__pyx_kp_b__105;
static PyObject *__pyx_kp_b__107;
static PyObject *__pyx_kp_b__109;
static PyObject *__pyx_kp_b__11;
static PyObject *__pyx_kp_b__111;
static PyObject *__pyx_kp_b__113;
static PyObject *__pyx_kp_b__115;
static PyObject *__pyx_kp_b__117;
static PyObject *__pyx_kp_b__119;
static PyObject *__pyx_kp_b__121;
static PyObject *__pyx_kp_b__123;
static PyObject *__pyx_kp_b__125;
static PyObject *__pyx_kp_b__127;
static PyObject *__pyx_kp_b__129;
static PyObject *__pyx_kp_b__13;
static PyObject *__pyx_kp_b__131;
static PyObject *__pyx_kp_b__133;
static PyObject *__pyx_kp_b__135;
static PyObject *__pyx_kp_b__137;
static PyObject *__pyx_kp_b__139;
static PyObject *__pyx_kp_b__141;
static PyObject *__pyx_kp_b__143;
static PyObject *__pyx_kp_b__146;
static PyObject *__pyx_kp_b__149;
static PyObject *__pyx_kp_b__15;
static PyObject *__pyx_kp_b__151;
static PyObject *__pyx_kp_b__153;
static PyObject *__pyx_kp_b__155;
static PyObject *__pyx_kp_b__157;
static PyObject *__pyx_kp_b__159;
static PyObject *__pyx_kp_b__161;
static PyObject *__pyx_kp_b__163;
static PyObject *__pyx_kp_b__165;
static PyObject *__pyx_kp_b__167;
static PyObject *__pyx_kp_b__169;
static PyObject *__pyx_kp_b__171;
static PyObject *__pyx_kp_b__173;
static PyObject *__pyx_kp_b__175;
static PyObject *__pyx_kp_b__177;
static PyObject *__pyx_kp_b__179;
static PyObject *__pyx_kp_b__181;
static PyObject *__pyx_kp_b__183;
static PyObject *__pyx_kp_b__186;
static PyObject *__pyx_n_s__188;
static PyObject *__pyx_kp_b__20;
static PyObject *__pyx_kp_b__22;
static PyObject *__pyx_kp_b__24;
static PyObject *__pyx_kp_b__27;
static PyObject *__pyx_kp_b__29;
static PyObject *__pyx_kp_b__3;
static PyObject *__pyx_kp_b__31;
static PyObject *__pyx_kp_b__33;
static PyObject *__pyx_kp_b__35;
static PyObject *__pyx_kp_b__37;
static PyObject *__pyx_kp_b__39;
static PyObject *__pyx_kp_b__41;
static PyObject *__pyx_kp_b__43;
static PyObject *__pyx_kp_b__45;
static PyObject *__pyx_kp_b__47;
static PyObject *__pyx_kp_b__49;
static PyObject *__pyx_kp_b__5;
static PyObject *__pyx_kp_b__51;
static PyObject *__pyx_kp_b__53;
static PyObject *__pyx_kp_b__55;
static PyObject *__pyx_kp_b__57;
static PyObject *__pyx_kp_b__59;
static PyObject *__pyx_kp_b__61;
static PyObject *__pyx_kp_b__63;
static PyObject *__pyx_kp_b__66;
static PyObject *__pyx_kp_b__68;
static PyObject *__pyx_kp_b__7;
static PyObject *__pyx_kp_b__70;
static PyObject *__pyx_kp_b__72;
static PyObject *__pyx_kp_b__74;
static PyObject *__pyx_kp_b__76;
static PyObject *__pyx_kp_b__78;
static PyObject *__pyx_kp_b__80;
static PyObject *__pyx_kp_b__82;
static PyObject *__pyx_kp_b__84;
static PyObject *__pyx_kp_b__86;
static PyObject *__pyx_kp_b__88;
static PyObject *__pyx_kp_b__9;
static PyObject *__pyx_kp_b__90;
static PyObject *__pyx_kp_b__92;
static PyObject *__pyx_kp_b__94;
static PyObject *__pyx_kp_b__96;
static PyObject *__pyx_kp_b__98;
static PyObject *__pyx_n_s_constants;
static PyObject *__pyx_kp_s_d;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_fields;
static PyObject *__pyx_kp_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_integer;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_neo4j_compat;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_pack_end_of_stream;
static PyObject *__pyx_n_s_pack_list_stream_header;
static PyObject *__pyx_n_s_pack_map_stream_header;
static PyObject *__pyx_n_s_pack_struct;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_kp_s_q;
static PyObject *__pyx_n_s_signature;
static PyObject *__pyx_n_s_stream;
static PyObject *__pyx_n_s_string;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_struct_pack;
static PyObject *__pyx_n_s_structure;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_s_utf_8;
static PyObject *__pyx_n_s_write;
static int __pyx_pf_5neo4j_10packstream_7_packer_6Packer___cinit__(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_stream); /* proto */
static PyObject *__pyx_pf_5neo4j_10packstream_7_packer_6Packer_2pack(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5neo4j_10packstream_7_packer_6Packer_4pack_list_stream_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5neo4j_10packstream_7_packer_6Packer_6pack_map_stream_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5neo4j_10packstream_7_packer_6Packer_8pack_struct(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_signature, PyObject *__pyx_v_fields); /* proto */
static PyObject *__pyx_pf_5neo4j_10packstream_7_packer_6Packer_10pack_end_of_stream(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5neo4j_10packstream_7_packer_6Packer_14supports_bytes___get__(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self); /* proto */
static int __pyx_pf_5neo4j_10packstream_7_packer_6Packer_14supports_bytes_2__set__(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_tp_new_5neo4j_10packstream_7_packer_Packer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_128;
static PyObject *__pyx_int_256;
static PyObject *__pyx_int_32768;
static PyObject *__pyx_int_65536;
static PyObject *__pyx_int_2147483648;
static PyObject *__pyx_int_4294967296;
static PyObject *__pyx_int_9223372036854775808;
static PyObject *__pyx_int_neg_16;
static PyObject *__pyx_int_neg_128;
static PyObject *__pyx_int_neg_32768;
static PyObject *__pyx_int_neg_2147483648;
static PyObject *__pyx_int_neg_9223372036854775808;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__75;
static PyObject *__pyx_tuple__77;
static PyObject *__pyx_tuple__79;
static PyObject *__pyx_tuple__81;
static PyObject *__pyx_tuple__83;
static PyObject *__pyx_tuple__85;
static PyObject *__pyx_tuple__87;
static PyObject *__pyx_tuple__89;
static PyObject *__pyx_tuple__91;
static PyObject *__pyx_tuple__93;
static PyObject *__pyx_tuple__95;
static PyObject *__pyx_tuple__97;
static PyObject *__pyx_tuple__99;
static PyObject *__pyx_tuple__101;
static PyObject *__pyx_tuple__103;
static PyObject *__pyx_tuple__104;
static PyObject *__pyx_tuple__106;
static PyObject *__pyx_tuple__108;
static PyObject *__pyx_tuple__110;
static PyObject *__pyx_tuple__112;
static PyObject *__pyx_tuple__114;
static PyObject *__pyx_tuple__116;
static PyObject *__pyx_tuple__118;
static PyObject *__pyx_tuple__120;
static PyObject *__pyx_tuple__122;
static PyObject *__pyx_tuple__124;
static PyObject *__pyx_tuple__126;
static PyObject *__pyx_tuple__128;
static PyObject *__pyx_tuple__130;
static PyObject *__pyx_tuple__132;
static PyObject *__pyx_tuple__134;
static PyObject *__pyx_tuple__136;
static PyObject *__pyx_tuple__138;
static PyObject *__pyx_tuple__140;
static PyObject *__pyx_tuple__142;
static PyObject *__pyx_tuple__144;
static PyObject *__pyx_tuple__145;
static PyObject *__pyx_tuple__147;
static PyObject *__pyx_tuple__148;
static PyObject *__pyx_tuple__150;
static PyObject *__pyx_tuple__152;
static PyObject *__pyx_tuple__154;
static PyObject *__pyx_tuple__156;
static PyObject *__pyx_tuple__158;
static PyObject *__pyx_tuple__160;
static PyObject *__pyx_tuple__162;
static PyObject *__pyx_tuple__164;
static PyObject *__pyx_tuple__166;
static PyObject *__pyx_tuple__168;
static PyObject *__pyx_tuple__170;
static PyObject *__pyx_tuple__172;
static PyObject *__pyx_tuple__174;
static PyObject *__pyx_tuple__176;
static PyObject *__pyx_tuple__178;
static PyObject *__pyx_tuple__180;
static PyObject *__pyx_tuple__182;
static PyObject *__pyx_tuple__184;
static PyObject *__pyx_tuple__185;
static PyObject *__pyx_tuple__187;

/* "neo4j/packstream/_packer.pyx":41
 *     cdef _write
 * 
 *     def __cinit__(self, stream):             # <<<<<<<<<<<<<<
 *         self.supports_bytes = False
 *         self.stream = stream
 */

/* Python wrapper */
static int __pyx_pw_5neo4j_10packstream_7_packer_6Packer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5neo4j_10packstream_7_packer_6Packer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_stream = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_stream,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_stream)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 41, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_stream = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 41, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5neo4j_10packstream_7_packer_6Packer___cinit__(((struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self), __pyx_v_stream);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5neo4j_10packstream_7_packer_6Packer___cinit__(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "neo4j/packstream/_packer.pyx":42
 * 
 *     def __cinit__(self, stream):
 *         self.supports_bytes = False             # <<<<<<<<<<<<<<
 *         self.stream = stream
 *         self._write = self.stream.write
 */
  __pyx_v_self->supports_bytes = 0;

  /* "neo4j/packstream/_packer.pyx":43
 *     def __cinit__(self, stream):
 *         self.supports_bytes = False
 *         self.stream = stream             # <<<<<<<<<<<<<<
 *         self._write = self.stream.write
 * 
 */
  __Pyx_INCREF(__pyx_v_stream);
  __Pyx_GIVEREF(__pyx_v_stream);
  __Pyx_GOTREF(__pyx_v_self->stream);
  __Pyx_DECREF(__pyx_v_self->stream);
  __pyx_v_self->stream = __pyx_v_stream;

  /* "neo4j/packstream/_packer.pyx":44
 *         self.supports_bytes = False
 *         self.stream = stream
 *         self._write = self.stream.write             # <<<<<<<<<<<<<<
 * 
 *     cdef pack_raw(self, data):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->stream, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_write);
  __Pyx_DECREF(__pyx_v_self->_write);
  __pyx_v_self->_write = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":41
 *     cdef _write
 * 
 *     def __cinit__(self, stream):             # <<<<<<<<<<<<<<
 *         self.supports_bytes = False
 *         self.stream = stream
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "neo4j/packstream/_packer.pyx":46
 *         self._write = self.stream.write
 * 
 *     cdef pack_raw(self, data):             # <<<<<<<<<<<<<<
 *         self._write(data)
 * 
 */

static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_raw(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("pack_raw", 0);

  /* "neo4j/packstream/_packer.pyx":47
 * 
 *     cdef pack_raw(self, data):
 *         self._write(data)             # <<<<<<<<<<<<<<
 * 
 *     cpdef pack(self, value):
 */
  __Pyx_INCREF(__pyx_v_self->_write);
  __pyx_t_2 = __pyx_v_self->_write; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_data};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_data};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 47, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_INCREF(__pyx_v_data);
      __Pyx_GIVEREF(__pyx_v_data);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_data);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":46
 *         self._write = self.stream.write
 * 
 *     cdef pack_raw(self, data):             # <<<<<<<<<<<<<<
 *         self._write(data)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_raw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "neo4j/packstream/_packer.pyx":49
 *         self._write(data)
 * 
 *     cpdef pack(self, value):             # <<<<<<<<<<<<<<
 *         return self._pack(value)
 * 
 */

static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_3pack(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_value, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("pack", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5neo4j_10packstream_7_packer_6Packer_3pack)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_value};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_value};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 49, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_value);
          __Pyx_GIVEREF(__pyx_v_value);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_value);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "neo4j/packstream/_packer.pyx":50
 * 
 *     cpdef pack(self, value):
 *         return self._pack(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _pack(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->_pack(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "neo4j/packstream/_packer.pyx":49
 *         self._write(data)
 * 
 *     cpdef pack(self, value):             # <<<<<<<<<<<<<<
 *         return self._pack(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_3pack(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_3pack(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack (wrapper)", 0);
  __pyx_r = __pyx_pf_5neo4j_10packstream_7_packer_6Packer_2pack(((struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5neo4j_10packstream_7_packer_6Packer_2pack(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pack", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5neo4j_10packstream_7_packer_6Packer_pack(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "neo4j/packstream/_packer.pyx":52
 *         return self._pack(value)
 * 
 *     cdef _pack(self, value):             # <<<<<<<<<<<<<<
 *         write = self._write
 * 
 */

static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer__pack(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_write = NULL;
  PyObject *__pyx_v_value_bytes = NULL;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_signature = NULL;
  PyObject *__pyx_v_fields = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  __Pyx_RefNannySetupContext("_pack", 0);

  /* "neo4j/packstream/_packer.pyx":53
 * 
 *     cdef _pack(self, value):
 *         write = self._write             # <<<<<<<<<<<<<<
 * 
 *         # None
 */
  __pyx_t_1 = __pyx_v_self->_write;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_write = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":56
 * 
 *         # None
 *         if value is None:             # <<<<<<<<<<<<<<
 *             write(b"\xC0")  # NULL
 * 
 */
  __pyx_t_2 = (__pyx_v_value == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "neo4j/packstream/_packer.pyx":57
 *         # None
 *         if value is None:
 *             write(b"\xC0")  # NULL             # <<<<<<<<<<<<<<
 * 
 *         # Boolean
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":56
 * 
 *         # None
 *         if value is None:             # <<<<<<<<<<<<<<
 *             write(b"\xC0")  # NULL
 * 
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":60
 * 
 *         # Boolean
 *         elif value is True:             # <<<<<<<<<<<<<<
 *             write(b"\xC3")
 *         elif value is False:
 */
  __pyx_t_3 = (__pyx_v_value == Py_True);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":61
 *         # Boolean
 *         elif value is True:
 *             write(b"\xC3")             # <<<<<<<<<<<<<<
 *         elif value is False:
 *             write(b"\xC2")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":60
 * 
 *         # Boolean
 *         elif value is True:             # <<<<<<<<<<<<<<
 *             write(b"\xC3")
 *         elif value is False:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":62
 *         elif value is True:
 *             write(b"\xC3")
 *         elif value is False:             # <<<<<<<<<<<<<<
 *             write(b"\xC2")
 * 
 */
  __pyx_t_2 = (__pyx_v_value == Py_False);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "neo4j/packstream/_packer.pyx":63
 *             write(b"\xC3")
 *         elif value is False:
 *             write(b"\xC2")             # <<<<<<<<<<<<<<
 * 
 *         # Float (only double precision is supported)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":62
 *         elif value is True:
 *             write(b"\xC3")
 *         elif value is False:             # <<<<<<<<<<<<<<
 *             write(b"\xC2")
 * 
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":66
 * 
 *         # Float (only double precision is supported)
 *         elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *             write(b"\xC1")
 *             write(struct_pack(">d", value))
 */
  __pyx_t_3 = PyFloat_Check(__pyx_v_value); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":67
 *         # Float (only double precision is supported)
 *         elif isinstance(value, float):
 *             write(b"\xC1")             # <<<<<<<<<<<<<<
 *             write(struct_pack(">d", value))
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":68
 *         elif isinstance(value, float):
 *             write(b"\xC1")
 *             write(struct_pack(">d", value))             # <<<<<<<<<<<<<<
 * 
 *         # Integer
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_struct_pack); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_kp_s_d, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 68, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_kp_s_d, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 68, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 68, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_d);
      __Pyx_GIVEREF(__pyx_kp_s_d);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_kp_s_d);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 68, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_5 = __pyx_v_write; __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (!__pyx_t_8) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 68, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8); __pyx_t_8 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":66
 * 
 *         # Float (only double precision is supported)
 *         elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *             write(b"\xC1")
 *             write(struct_pack(">d", value))
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":71
 * 
 *         # Integer
 *         elif isinstance(value, integer):             # <<<<<<<<<<<<<<
 *             if -0x10 <= value < 0x80:
 *                 write(PACKED_UINT_8[value % 0x100])
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_integer); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "neo4j/packstream/_packer.pyx":72
 *         # Integer
 *         elif isinstance(value, integer):
 *             if -0x10 <= value < 0x80:             # <<<<<<<<<<<<<<
 *                 write(PACKED_UINT_8[value % 0x100])
 *             elif -0x80 <= value < -0x10:
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_int_neg_16, __pyx_v_value, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
    if (__Pyx_PyObject_IsTrue(__pyx_t_1)) {
      __Pyx_DECREF(__pyx_t_1);
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_value, __pyx_int_128, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 72, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "neo4j/packstream/_packer.pyx":73
 *         elif isinstance(value, integer):
 *             if -0x10 <= value < 0x80:
 *                 write(PACKED_UINT_8[value % 0x100])             # <<<<<<<<<<<<<<
 *             elif -0x80 <= value < -0x10:
 *                 write(b"\xC8")
 */
      __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 73, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyInt_RemainderObjC(__pyx_v_value, __pyx_int_256, 0x100, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 73, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = PyObject_GetItem(__pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 73, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_v_write);
      __pyx_t_6 = __pyx_v_write; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 73, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "neo4j/packstream/_packer.pyx":72
 *         # Integer
 *         elif isinstance(value, integer):
 *             if -0x10 <= value < 0x80:             # <<<<<<<<<<<<<<
 *                 write(PACKED_UINT_8[value % 0x100])
 *             elif -0x80 <= value < -0x10:
 */
      goto __pyx_L4;
    }

    /* "neo4j/packstream/_packer.pyx":74
 *             if -0x10 <= value < 0x80:
 *                 write(PACKED_UINT_8[value % 0x100])
 *             elif -0x80 <= value < -0x10:             # <<<<<<<<<<<<<<
 *                 write(b"\xC8")
 *                 write(PACKED_UINT_8[value % 0x100])
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_int_neg_128, __pyx_v_value, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
    if (__Pyx_PyObject_IsTrue(__pyx_t_1)) {
      __Pyx_DECREF(__pyx_t_1);
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_value, __pyx_int_neg_16, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "neo4j/packstream/_packer.pyx":75
 *                 write(PACKED_UINT_8[value % 0x100])
 *             elif -0x80 <= value < -0x10:
 *                 write(b"\xC8")             # <<<<<<<<<<<<<<
 *                 write(PACKED_UINT_8[value % 0x100])
 *             elif -0x8000 <= value < 0x8000:
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 75, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "neo4j/packstream/_packer.pyx":76
 *             elif -0x80 <= value < -0x10:
 *                 write(b"\xC8")
 *                 write(PACKED_UINT_8[value % 0x100])             # <<<<<<<<<<<<<<
 *             elif -0x8000 <= value < 0x8000:
 *                 write(b"\xC9")
 */
      __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = __Pyx_PyInt_RemainderObjC(__pyx_v_value, __pyx_int_256, 0x100, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = PyObject_GetItem(__pyx_t_6, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_INCREF(__pyx_v_write);
      __pyx_t_8 = __pyx_v_write; __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      if (!__pyx_t_6) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_t_4};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_t_4};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 76, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "neo4j/packstream/_packer.pyx":74
 *             if -0x10 <= value < 0x80:
 *                 write(PACKED_UINT_8[value % 0x100])
 *             elif -0x80 <= value < -0x10:             # <<<<<<<<<<<<<<
 *                 write(b"\xC8")
 *                 write(PACKED_UINT_8[value % 0x100])
 */
      goto __pyx_L4;
    }

    /* "neo4j/packstream/_packer.pyx":77
 *                 write(b"\xC8")
 *                 write(PACKED_UINT_8[value % 0x100])
 *             elif -0x8000 <= value < 0x8000:             # <<<<<<<<<<<<<<
 *                 write(b"\xC9")
 *                 write(PACKED_UINT_16[value % 0x10000])
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_int_neg_32768, __pyx_v_value, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)
    if (__Pyx_PyObject_IsTrue(__pyx_t_1)) {
      __Pyx_DECREF(__pyx_t_1);
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_value, __pyx_int_32768, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "neo4j/packstream/_packer.pyx":78
 *                 write(PACKED_UINT_8[value % 0x100])
 *             elif -0x8000 <= value < 0x8000:
 *                 write(b"\xC9")             # <<<<<<<<<<<<<<
 *                 write(PACKED_UINT_16[value % 0x10000])
 *             elif -0x80000000 <= value < 0x80000000:
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "neo4j/packstream/_packer.pyx":79
 *             elif -0x8000 <= value < 0x8000:
 *                 write(b"\xC9")
 *                 write(PACKED_UINT_16[value % 0x10000])             # <<<<<<<<<<<<<<
 *             elif -0x80000000 <= value < 0x80000000:
 *                 write(b"\xCA")
 */
      __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_16); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_5 = __Pyx_PyInt_RemainderObjC(__pyx_v_value, __pyx_int_65536, 0x10000, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PyObject_GetItem(__pyx_t_8, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_v_write);
      __pyx_t_5 = __pyx_v_write; __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_8) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_4};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_4};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8); __pyx_t_8 = NULL;
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "neo4j/packstream/_packer.pyx":77
 *                 write(b"\xC8")
 *                 write(PACKED_UINT_8[value % 0x100])
 *             elif -0x8000 <= value < 0x8000:             # <<<<<<<<<<<<<<
 *                 write(b"\xC9")
 *                 write(PACKED_UINT_16[value % 0x10000])
 */
      goto __pyx_L4;
    }

    /* "neo4j/packstream/_packer.pyx":80
 *                 write(b"\xC9")
 *                 write(PACKED_UINT_16[value % 0x10000])
 *             elif -0x80000000 <= value < 0x80000000:             # <<<<<<<<<<<<<<
 *                 write(b"\xCA")
 *                 write(struct_pack(">i", value))
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_int_neg_2147483648, __pyx_v_value, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
    if (__Pyx_PyObject_IsTrue(__pyx_t_1)) {
      __Pyx_DECREF(__pyx_t_1);
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_value, __pyx_int_2147483648, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "neo4j/packstream/_packer.pyx":81
 *                 write(PACKED_UINT_16[value % 0x10000])
 *             elif -0x80000000 <= value < 0x80000000:
 *                 write(b"\xCA")             # <<<<<<<<<<<<<<
 *                 write(struct_pack(">i", value))
 *             elif INT64_LO <= value < INT64_HI:
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "neo4j/packstream/_packer.pyx":82
 *             elif -0x80000000 <= value < 0x80000000:
 *                 write(b"\xCA")
 *                 write(struct_pack(">i", value))             # <<<<<<<<<<<<<<
 *             elif INT64_LO <= value < INT64_HI:
 *                 write(b"\xCB")
 */
      __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_struct_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 82, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_s_i, __pyx_v_value};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_s_i, __pyx_v_value};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(__pyx_kp_s_i);
        __Pyx_GIVEREF(__pyx_kp_s_i);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_kp_s_i);
        __Pyx_INCREF(__pyx_v_value);
        __Pyx_GIVEREF(__pyx_v_value);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_v_write);
      __pyx_t_6 = __pyx_v_write; __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      if (!__pyx_t_8) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_5};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_5};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else
        #endif
        {
          __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 82, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8); __pyx_t_8 = NULL;
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_5);
          __pyx_t_5 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "neo4j/packstream/_packer.pyx":80
 *                 write(b"\xC9")
 *                 write(PACKED_UINT_16[value % 0x10000])
 *             elif -0x80000000 <= value < 0x80000000:             # <<<<<<<<<<<<<<
 *                 write(b"\xCA")
 *                 write(struct_pack(">i", value))
 */
      goto __pyx_L4;
    }

    /* "neo4j/packstream/_packer.pyx":83
 *                 write(b"\xCA")
 *                 write(struct_pack(">i", value))
 *             elif INT64_LO <= value < INT64_HI:             # <<<<<<<<<<<<<<
 *                 write(b"\xCB")
 *                 write(struct_pack(">q", value))
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_5neo4j_10packstream_7_packer_INT64_LO, __pyx_v_value, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
    if (__Pyx_PyObject_IsTrue(__pyx_t_1)) {
      __Pyx_DECREF(__pyx_t_1);
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_value, __pyx_v_5neo4j_10packstream_7_packer_INT64_HI, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "neo4j/packstream/_packer.pyx":84
 *                 write(struct_pack(">i", value))
 *             elif INT64_LO <= value < INT64_HI:
 *                 write(b"\xCB")             # <<<<<<<<<<<<<<
 *                 write(struct_pack(">q", value))
 *             else:
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "neo4j/packstream/_packer.pyx":85
 *             elif INT64_LO <= value < INT64_HI:
 *                 write(b"\xCB")
 *                 write(struct_pack(">q", value))             # <<<<<<<<<<<<<<
 *             else:
 *                 raise OverflowError("Integer %s out of range" % value)
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_struct_pack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 85, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s_q, __pyx_v_value};
        __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 85, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s_q, __pyx_v_value};
        __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 85, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 85, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_INCREF(__pyx_kp_s_q);
        __Pyx_GIVEREF(__pyx_kp_s_q);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_kp_s_q);
        __Pyx_INCREF(__pyx_v_value);
        __Pyx_GIVEREF(__pyx_v_value);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 85, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_INCREF(__pyx_v_write);
      __pyx_t_4 = __pyx_v_write; __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_8) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_6};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_6};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 85, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8); __pyx_t_8 = NULL;
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_6);
          __pyx_t_6 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "neo4j/packstream/_packer.pyx":83
 *                 write(b"\xCA")
 *                 write(struct_pack(">i", value))
 *             elif INT64_LO <= value < INT64_HI:             # <<<<<<<<<<<<<<
 *                 write(b"\xCB")
 *                 write(struct_pack(">q", value))
 */
      goto __pyx_L4;
    }

    /* "neo4j/packstream/_packer.pyx":87
 *                 write(struct_pack(">q", value))
 *             else:
 *                 raise OverflowError("Integer %s out of range" % value)             # <<<<<<<<<<<<<<
 * 
 *         # String
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Integer_s_out_of_range, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 87, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_OverflowError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 87, __pyx_L1_error)
    }
    __pyx_L4:;

    /* "neo4j/packstream/_packer.pyx":71
 * 
 *         # Integer
 *         elif isinstance(value, integer):             # <<<<<<<<<<<<<<
 *             if -0x10 <= value < 0x80:
 *                 write(PACKED_UINT_8[value % 0x100])
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":90
 * 
 *         # String
 *         elif isinstance(value, string):             # <<<<<<<<<<<<<<
 *             if isinstance(value, bytes):
 *                 value_bytes = value
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":91
 *         # String
 *         elif isinstance(value, string):
 *             if isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 value_bytes = value
 *             else:
 */
    __pyx_t_2 = PyBytes_Check(__pyx_v_value); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {

      /* "neo4j/packstream/_packer.pyx":92
 *         elif isinstance(value, string):
 *             if isinstance(value, bytes):
 *                 value_bytes = value             # <<<<<<<<<<<<<<
 *             else:
 *                 value_bytes = value.encode("utf-8")
 */
      __Pyx_INCREF(__pyx_v_value);
      __pyx_v_value_bytes = __pyx_v_value;

      /* "neo4j/packstream/_packer.pyx":91
 *         # String
 *         elif isinstance(value, string):
 *             if isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 value_bytes = value
 *             else:
 */
      goto __pyx_L5;
    }

    /* "neo4j/packstream/_packer.pyx":94
 *                 value_bytes = value
 *             else:
 *                 value_bytes = value.encode("utf-8")             # <<<<<<<<<<<<<<
 *             self.pack_string_header(len(value_bytes))
 *             self.pack_raw(value_bytes)
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 94, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_value_bytes = __pyx_t_4;
      __pyx_t_4 = 0;
    }
    __pyx_L5:;

    /* "neo4j/packstream/_packer.pyx":95
 *             else:
 *                 value_bytes = value.encode("utf-8")
 *             self.pack_string_header(len(value_bytes))             # <<<<<<<<<<<<<<
 *             self.pack_raw(value_bytes)
 * 
 */
    __pyx_t_9 = PyObject_Length(__pyx_v_value_bytes); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 95, __pyx_L1_error)
    __pyx_t_4 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->pack_string_header(__pyx_v_self, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "neo4j/packstream/_packer.pyx":96
 *                 value_bytes = value.encode("utf-8")
 *             self.pack_string_header(len(value_bytes))
 *             self.pack_raw(value_bytes)             # <<<<<<<<<<<<<<
 * 
 *         # Bytes (deliberately listed after String since in
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->pack_raw(__pyx_v_self, __pyx_v_value_bytes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "neo4j/packstream/_packer.pyx":90
 * 
 *         # String
 *         elif isinstance(value, string):             # <<<<<<<<<<<<<<
 *             if isinstance(value, bytes):
 *                 value_bytes = value
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":100
 *         # Bytes (deliberately listed after String since in
 *         # Python 2, bytes should be treated as a String)
 *         elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *             self.pack_bytes_header(len(value))
 *             self.pack_raw(value)
 */
  __pyx_t_3 = PyBytes_Check(__pyx_v_value); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":101
 *         # Python 2, bytes should be treated as a String)
 *         elif isinstance(value, bytes):
 *             self.pack_bytes_header(len(value))             # <<<<<<<<<<<<<<
 *             self.pack_raw(value)
 *         elif isinstance(value, bytearray):
 */
    __pyx_t_9 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 101, __pyx_L1_error)
    __pyx_t_4 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->pack_bytes_header(__pyx_v_self, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "neo4j/packstream/_packer.pyx":102
 *         elif isinstance(value, bytes):
 *             self.pack_bytes_header(len(value))
 *             self.pack_raw(value)             # <<<<<<<<<<<<<<
 *         elif isinstance(value, bytearray):
 *             self.pack_bytes_header(len(value))
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->pack_raw(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "neo4j/packstream/_packer.pyx":100
 *         # Bytes (deliberately listed after String since in
 *         # Python 2, bytes should be treated as a String)
 *         elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *             self.pack_bytes_header(len(value))
 *             self.pack_raw(value)
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":103
 *             self.pack_bytes_header(len(value))
 *             self.pack_raw(value)
 *         elif isinstance(value, bytearray):             # <<<<<<<<<<<<<<
 *             self.pack_bytes_header(len(value))
 *             self.pack_raw(bytes(value))
 */
  __pyx_t_2 = PyByteArray_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "neo4j/packstream/_packer.pyx":104
 *             self.pack_raw(value)
 *         elif isinstance(value, bytearray):
 *             self.pack_bytes_header(len(value))             # <<<<<<<<<<<<<<
 *             self.pack_raw(bytes(value))
 * 
 */
    __pyx_t_9 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 104, __pyx_L1_error)
    __pyx_t_4 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->pack_bytes_header(__pyx_v_self, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "neo4j/packstream/_packer.pyx":105
 *         elif isinstance(value, bytearray):
 *             self.pack_bytes_header(len(value))
 *             self.pack_raw(bytes(value))             # <<<<<<<<<<<<<<
 * 
 *         # List
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_value);
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)(&PyBytes_Type)), __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->pack_raw(__pyx_v_self, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "neo4j/packstream/_packer.pyx":103
 *             self.pack_bytes_header(len(value))
 *             self.pack_raw(value)
 *         elif isinstance(value, bytearray):             # <<<<<<<<<<<<<<
 *             self.pack_bytes_header(len(value))
 *             self.pack_raw(bytes(value))
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":108
 * 
 *         # List
 *         elif isinstance(value, list):             # <<<<<<<<<<<<<<
 *             self.pack_list_header(len(value))
 *             for item in value:
 */
  __pyx_t_3 = PyList_Check(__pyx_v_value); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":109
 *         # List
 *         elif isinstance(value, list):
 *             self.pack_list_header(len(value))             # <<<<<<<<<<<<<<
 *             for item in value:
 *                 self._pack(item)
 */
    __pyx_t_9 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 109, __pyx_L1_error)
    __pyx_t_4 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->pack_list_header(__pyx_v_self, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "neo4j/packstream/_packer.pyx":110
 *         elif isinstance(value, list):
 *             self.pack_list_header(len(value))
 *             for item in value:             # <<<<<<<<<<<<<<
 *                 self._pack(item)
 * 
 */
    if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
      __pyx_t_4 = __pyx_v_value; __Pyx_INCREF(__pyx_t_4); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_10 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 110, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 110, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 110, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_10(__pyx_t_4);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 110, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "neo4j/packstream/_packer.pyx":111
 *             self.pack_list_header(len(value))
 *             for item in value:
 *                 self._pack(item)             # <<<<<<<<<<<<<<
 * 
 *         # Map
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->_pack(__pyx_v_self, __pyx_v_item); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "neo4j/packstream/_packer.pyx":110
 *         elif isinstance(value, list):
 *             self.pack_list_header(len(value))
 *             for item in value:             # <<<<<<<<<<<<<<
 *                 self._pack(item)
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "neo4j/packstream/_packer.pyx":108
 * 
 *         # List
 *         elif isinstance(value, list):             # <<<<<<<<<<<<<<
 *             self.pack_list_header(len(value))
 *             for item in value:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":114
 * 
 *         # Map
 *         elif isinstance(value, dict):             # <<<<<<<<<<<<<<
 *             self.pack_map_header(len(value))
 *             for key, item in value.items():
 */
  __pyx_t_2 = PyDict_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "neo4j/packstream/_packer.pyx":115
 *         # Map
 *         elif isinstance(value, dict):
 *             self.pack_map_header(len(value))             # <<<<<<<<<<<<<<
 *             for key, item in value.items():
 *                 self._pack(key)
 */
    __pyx_t_9 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 115, __pyx_L1_error)
    __pyx_t_4 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->pack_map_header(__pyx_v_self, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "neo4j/packstream/_packer.pyx":116
 *         elif isinstance(value, dict):
 *             self.pack_map_header(len(value))
 *             for key, item in value.items():             # <<<<<<<<<<<<<<
 *                 self._pack(key)
 *                 self._pack(item)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 116, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 116, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_1 = __pyx_t_4; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 116, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 116, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 116, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 116, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 116, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 116, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_10(__pyx_t_1);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 116, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        #if !CYTHON_COMPILING_IN_PYPY
        Py_ssize_t size = Py_SIZE(sequence);
        #else
        Py_ssize_t size = PySequence_Size(sequence);
        #endif
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 116, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 116, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 116, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 116, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = Py_TYPE(__pyx_t_8)->tp_iternext;
        index = 0; __pyx_t_5 = __pyx_t_11(__pyx_t_8); if (unlikely(!__pyx_t_5)) goto __pyx_L10_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_6 = __pyx_t_11(__pyx_t_8); if (unlikely(!__pyx_t_6)) goto __pyx_L10_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_8), 2) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
        __pyx_t_11 = NULL;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L11_unpacking_done;
        __pyx_L10_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_11 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 116, __pyx_L1_error)
        __pyx_L11_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "neo4j/packstream/_packer.pyx":117
 *             self.pack_map_header(len(value))
 *             for key, item in value.items():
 *                 self._pack(key)             # <<<<<<<<<<<<<<
 *                 self._pack(item)
 * 
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->_pack(__pyx_v_self, __pyx_v_key); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "neo4j/packstream/_packer.pyx":118
 *             for key, item in value.items():
 *                 self._pack(key)
 *                 self._pack(item)             # <<<<<<<<<<<<<<
 * 
 *         # Structure
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->_pack(__pyx_v_self, __pyx_v_item); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "neo4j/packstream/_packer.pyx":116
 *         elif isinstance(value, dict):
 *             self.pack_map_header(len(value))
 *             for key, item in value.items():             # <<<<<<<<<<<<<<
 *                 self._pack(key)
 *                 self._pack(item)
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":114
 * 
 *         # Map
 *         elif isinstance(value, dict):             # <<<<<<<<<<<<<<
 *             self.pack_map_header(len(value))
 *             for key, item in value.items():
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":121
 * 
 *         # Structure
 *         elif isinstance(value, (Structure, tuple)):             # <<<<<<<<<<<<<<
 *             try:
 *                 signature, fields = value
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_Structure); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
  __pyx_t_12 = (__pyx_t_2 != 0);
  if (!__pyx_t_12) {
  } else {
    __pyx_t_3 = __pyx_t_12;
    goto __pyx_L12_bool_binop_done;
  }
  __pyx_t_12 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_2 = (__pyx_t_12 != 0);
  __pyx_t_3 = __pyx_t_2;
  __pyx_L12_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":122
 *         # Structure
 *         elif isinstance(value, (Structure, tuple)):
 *             try:             # <<<<<<<<<<<<<<
 *                 signature, fields = value
 *             except ValueError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      /*try:*/ {

        /* "neo4j/packstream/_packer.pyx":123
 *         elif isinstance(value, (Structure, tuple)):
 *             try:
 *                 signature, fields = value             # <<<<<<<<<<<<<<
 *             except ValueError:
 *                 raise ValueError("Structures require a 2-tuple of (signature, fields)")
 */
        if ((likely(PyTuple_CheckExact(__pyx_v_value))) || (PyList_CheckExact(__pyx_v_value))) {
          PyObject* sequence = __pyx_v_value;
          #if !CYTHON_COMPILING_IN_PYPY
          Py_ssize_t size = Py_SIZE(sequence);
          #else
          Py_ssize_t size = PySequence_Size(sequence);
          #endif
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 123, __pyx_L14_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          #else
          __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          Py_ssize_t index = -1;
          __pyx_t_6 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_11 = Py_TYPE(__pyx_t_6)->tp_iternext;
          index = 0; __pyx_t_1 = __pyx_t_11(__pyx_t_6); if (unlikely(!__pyx_t_1)) goto __pyx_L22_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_1);
          index = 1; __pyx_t_4 = __pyx_t_11(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L22_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_4);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_6), 2) < 0) __PYX_ERR(0, 123, __pyx_L14_error)
          __pyx_t_11 = NULL;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          goto __pyx_L23_unpacking_done;
          __pyx_L22_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_11 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 123, __pyx_L14_error)
          __pyx_L23_unpacking_done:;
        }
        __pyx_v_signature = __pyx_t_1;
        __pyx_t_1 = 0;
        __pyx_v_fields = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "neo4j/packstream/_packer.pyx":122
 *         # Structure
 *         elif isinstance(value, (Structure, tuple)):
 *             try:             # <<<<<<<<<<<<<<
 *                 signature, fields = value
 *             except ValueError:
 */
      }

      /* "neo4j/packstream/_packer.pyx":127
 *                 raise ValueError("Structures require a 2-tuple of (signature, fields)")
 *             else:
 *                 self.pack_struct(signature, fields)             # <<<<<<<<<<<<<<
 * 
 *         # Other
 */
      /*else:*/ {
        if (!(likely(PyBytes_CheckExact(__pyx_v_signature))||((__pyx_v_signature) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_signature)->tp_name), 0))) __PYX_ERR(0, 127, __pyx_L16_except_error)
        if (!(likely(PyTuple_CheckExact(__pyx_v_fields))||((__pyx_v_fields) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_fields)->tp_name), 0))) __PYX_ERR(0, 127, __pyx_L16_except_error)
        __pyx_t_4 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->pack_struct(__pyx_v_self, ((PyObject*)__pyx_v_signature), ((PyObject*)__pyx_v_fields), 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L16_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      goto __pyx_L21_try_end;
      __pyx_L14_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "neo4j/packstream/_packer.pyx":124
 *             try:
 *                 signature, fields = value
 *             except ValueError:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Structures require a 2-tuple of (signature, fields)")
 *             else:
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
      if (__pyx_t_7) {
        __Pyx_AddTraceback("neo4j.packstream._packer.Packer._pack", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_6) < 0) __PYX_ERR(0, 124, __pyx_L16_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_6);

        /* "neo4j/packstream/_packer.pyx":125
 *                 signature, fields = value
 *             except ValueError:
 *                 raise ValueError("Structures require a 2-tuple of (signature, fields)")             # <<<<<<<<<<<<<<
 *             else:
 *                 self.pack_struct(signature, fields)
 */
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 125, __pyx_L16_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_Raise(__pyx_t_5, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __PYX_ERR(0, 125, __pyx_L16_except_error)
      }
      goto __pyx_L16_except_error;
      __pyx_L16_except_error:;

      /* "neo4j/packstream/_packer.pyx":122
 *         # Structure
 *         elif isinstance(value, (Structure, tuple)):
 *             try:             # <<<<<<<<<<<<<<
 *                 signature, fields = value
 *             except ValueError:
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      goto __pyx_L1_error;
      __pyx_L21_try_end:;
    }

    /* "neo4j/packstream/_packer.pyx":121
 * 
 *         # Structure
 *         elif isinstance(value, (Structure, tuple)):             # <<<<<<<<<<<<<<
 *             try:
 *                 signature, fields = value
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":131
 *         # Other
 *         else:
 *             raise ValueError("Values of type %s are not supported" % type(value))             # <<<<<<<<<<<<<<
 * 
 *     cdef pack_bytes_header(self, int size):
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Values_of_type_s_are_not_support, ((PyObject *)Py_TYPE(__pyx_v_value))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 131, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "neo4j/packstream/_packer.pyx":52
 *         return self._pack(value)
 * 
 *     cdef _pack(self, value):             # <<<<<<<<<<<<<<
 *         write = self._write
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer._pack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_write);
  __Pyx_XDECREF(__pyx_v_value_bytes);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_signature);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "neo4j/packstream/_packer.pyx":133
 *             raise ValueError("Values of type %s are not supported" % type(value))
 * 
 *     cdef pack_bytes_header(self, int size):             # <<<<<<<<<<<<<<
 *         if not self.supports_bytes:
 *             raise TypeError("This PackSteam channel does not support BYTES (consider upgrading to Neo4j 3.2+)")
 */

static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_bytes_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_v_size) {
  PyObject *__pyx_v_write = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("pack_bytes_header", 0);

  /* "neo4j/packstream/_packer.pyx":134
 * 
 *     cdef pack_bytes_header(self, int size):
 *         if not self.supports_bytes:             # <<<<<<<<<<<<<<
 *             raise TypeError("This PackSteam channel does not support BYTES (consider upgrading to Neo4j 3.2+)")
 *         write = self._write
 */
  __pyx_t_1 = ((!(__pyx_v_self->supports_bytes != 0)) != 0);
  if (__pyx_t_1) {

    /* "neo4j/packstream/_packer.pyx":135
 *     cdef pack_bytes_header(self, int size):
 *         if not self.supports_bytes:
 *             raise TypeError("This PackSteam channel does not support BYTES (consider upgrading to Neo4j 3.2+)")             # <<<<<<<<<<<<<<
 *         write = self._write
 *         if size < 0x100:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 135, __pyx_L1_error)

    /* "neo4j/packstream/_packer.pyx":134
 * 
 *     cdef pack_bytes_header(self, int size):
 *         if not self.supports_bytes:             # <<<<<<<<<<<<<<
 *             raise TypeError("This PackSteam channel does not support BYTES (consider upgrading to Neo4j 3.2+)")
 *         write = self._write
 */
  }

  /* "neo4j/packstream/_packer.pyx":136
 *         if not self.supports_bytes:
 *             raise TypeError("This PackSteam channel does not support BYTES (consider upgrading to Neo4j 3.2+)")
 *         write = self._write             # <<<<<<<<<<<<<<
 *         if size < 0x100:
 *             write(b"\xCC")
 */
  __pyx_t_2 = __pyx_v_self->_write;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_write = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "neo4j/packstream/_packer.pyx":137
 *             raise TypeError("This PackSteam channel does not support BYTES (consider upgrading to Neo4j 3.2+)")
 *         write = self._write
 *         if size < 0x100:             # <<<<<<<<<<<<<<
 *             write(b"\xCC")
 *             write(PACKED_UINT_8[size])
 */
  __pyx_t_1 = ((__pyx_v_size < 0x100) != 0);
  if (__pyx_t_1) {

    /* "neo4j/packstream/_packer.pyx":138
 *         write = self._write
 *         if size < 0x100:
 *             write(b"\xCC")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "neo4j/packstream/_packer.pyx":139
 *         if size < 0x100:
 *             write(b"\xCC")
 *             write(PACKED_UINT_8[size])             # <<<<<<<<<<<<<<
 *         elif size < 0x10000:
 *             write(b"\xCD")
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_size, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_3 = __pyx_v_write; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 139, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "neo4j/packstream/_packer.pyx":137
 *             raise TypeError("This PackSteam channel does not support BYTES (consider upgrading to Neo4j 3.2+)")
 *         write = self._write
 *         if size < 0x100:             # <<<<<<<<<<<<<<
 *             write(b"\xCC")
 *             write(PACKED_UINT_8[size])
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":140
 *             write(b"\xCC")
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:             # <<<<<<<<<<<<<<
 *             write(b"\xCD")
 *             write(PACKED_UINT_16[size])
 */
  __pyx_t_1 = ((__pyx_v_size < 0x10000) != 0);
  if (__pyx_t_1) {

    /* "neo4j/packstream/_packer.pyx":141
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 *             write(b"\xCD")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "neo4j/packstream/_packer.pyx":142
 *         elif size < 0x10000:
 *             write(b"\xCD")
 *             write(PACKED_UINT_16[size])             # <<<<<<<<<<<<<<
 *         elif size < 0x100000000:
 *             write(b"\xCE")
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_size, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_3 = __pyx_v_write; __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_4) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_6};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_6};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 142, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "neo4j/packstream/_packer.pyx":140
 *             write(b"\xCC")
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:             # <<<<<<<<<<<<<<
 *             write(b"\xCD")
 *             write(PACKED_UINT_16[size])
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":143
 *             write(b"\xCD")
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:             # <<<<<<<<<<<<<<
 *             write(b"\xCE")
 *             write(struct_pack(">I", size))
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_4294967296, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "neo4j/packstream/_packer.pyx":144
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 *             write(b"\xCE")             # <<<<<<<<<<<<<<
 *             write(struct_pack(">I", size))
 *         else:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "neo4j/packstream/_packer.pyx":145
 *         elif size < 0x100000000:
 *             write(b"\xCE")
 *             write(struct_pack(">I", size))             # <<<<<<<<<<<<<<
 *         else:
 *             raise OverflowError("Bytes header size out of range")
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_struct_pack); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_s_I, __pyx_t_6};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_s_I, __pyx_t_6};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_I);
      __Pyx_GIVEREF(__pyx_kp_s_I);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_kp_s_I);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_5 = __pyx_v_write; __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (!__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 145, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_2};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 145, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_2};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 145, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 145, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8); __pyx_t_8 = NULL;
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 145, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "neo4j/packstream/_packer.pyx":143
 *             write(b"\xCD")
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:             # <<<<<<<<<<<<<<
 *             write(b"\xCE")
 *             write(struct_pack(">I", size))
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":147
 *             write(struct_pack(">I", size))
 *         else:
 *             raise OverflowError("Bytes header size out of range")             # <<<<<<<<<<<<<<
 * 
 *     cdef pack_string_header(self, int size):
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_OverflowError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 147, __pyx_L1_error)
  }
  __pyx_L4:;

  /* "neo4j/packstream/_packer.pyx":133
 *             raise ValueError("Values of type %s are not supported" % type(value))
 * 
 *     cdef pack_bytes_header(self, int size):             # <<<<<<<<<<<<<<
 *         if not self.supports_bytes:
 *             raise TypeError("This PackSteam channel does not support BYTES (consider upgrading to Neo4j 3.2+)")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_bytes_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_write);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "neo4j/packstream/_packer.pyx":149
 *             raise OverflowError("Bytes header size out of range")
 * 
 *     cdef pack_string_header(self, int size):             # <<<<<<<<<<<<<<
 *         write = self._write
 *         if size == 0x00:
 */

static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_string_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_v_size) {
  PyObject *__pyx_v_write = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("pack_string_header", 0);

  /* "neo4j/packstream/_packer.pyx":150
 * 
 *     cdef pack_string_header(self, int size):
 *         write = self._write             # <<<<<<<<<<<<<<
 *         if size == 0x00:
 *             write(b"\x80")
 */
  __pyx_t_1 = __pyx_v_self->_write;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_write = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":151
 *     cdef pack_string_header(self, int size):
 *         write = self._write
 *         if size == 0x00:             # <<<<<<<<<<<<<<
 *             write(b"\x80")
 *         elif size == 0x01:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x00) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":152
 *         write = self._write
 *         if size == 0x00:
 *             write(b"\x80")             # <<<<<<<<<<<<<<
 *         elif size == 0x01:
 *             write(b"\x81")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":151
 *     cdef pack_string_header(self, int size):
 *         write = self._write
 *         if size == 0x00:             # <<<<<<<<<<<<<<
 *             write(b"\x80")
 *         elif size == 0x01:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":153
 *         if size == 0x00:
 *             write(b"\x80")
 *         elif size == 0x01:             # <<<<<<<<<<<<<<
 *             write(b"\x81")
 *         elif size == 0x02:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x01) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":154
 *             write(b"\x80")
 *         elif size == 0x01:
 *             write(b"\x81")             # <<<<<<<<<<<<<<
 *         elif size == 0x02:
 *             write(b"\x82")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":153
 *         if size == 0x00:
 *             write(b"\x80")
 *         elif size == 0x01:             # <<<<<<<<<<<<<<
 *             write(b"\x81")
 *         elif size == 0x02:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":155
 *         elif size == 0x01:
 *             write(b"\x81")
 *         elif size == 0x02:             # <<<<<<<<<<<<<<
 *             write(b"\x82")
 *         elif size == 0x03:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x02) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":156
 *             write(b"\x81")
 *         elif size == 0x02:
 *             write(b"\x82")             # <<<<<<<<<<<<<<
 *         elif size == 0x03:
 *             write(b"\x83")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":155
 *         elif size == 0x01:
 *             write(b"\x81")
 *         elif size == 0x02:             # <<<<<<<<<<<<<<
 *             write(b"\x82")
 *         elif size == 0x03:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":157
 *         elif size == 0x02:
 *             write(b"\x82")
 *         elif size == 0x03:             # <<<<<<<<<<<<<<
 *             write(b"\x83")
 *         elif size == 0x04:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x03) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":158
 *             write(b"\x82")
 *         elif size == 0x03:
 *             write(b"\x83")             # <<<<<<<<<<<<<<
 *         elif size == 0x04:
 *             write(b"\x84")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":157
 *         elif size == 0x02:
 *             write(b"\x82")
 *         elif size == 0x03:             # <<<<<<<<<<<<<<
 *             write(b"\x83")
 *         elif size == 0x04:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":159
 *         elif size == 0x03:
 *             write(b"\x83")
 *         elif size == 0x04:             # <<<<<<<<<<<<<<
 *             write(b"\x84")
 *         elif size == 0x05:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x04) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":160
 *             write(b"\x83")
 *         elif size == 0x04:
 *             write(b"\x84")             # <<<<<<<<<<<<<<
 *         elif size == 0x05:
 *             write(b"\x85")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":159
 *         elif size == 0x03:
 *             write(b"\x83")
 *         elif size == 0x04:             # <<<<<<<<<<<<<<
 *             write(b"\x84")
 *         elif size == 0x05:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":161
 *         elif size == 0x04:
 *             write(b"\x84")
 *         elif size == 0x05:             # <<<<<<<<<<<<<<
 *             write(b"\x85")
 *         elif size == 0x06:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x05) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":162
 *             write(b"\x84")
 *         elif size == 0x05:
 *             write(b"\x85")             # <<<<<<<<<<<<<<
 *         elif size == 0x06:
 *             write(b"\x86")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":161
 *         elif size == 0x04:
 *             write(b"\x84")
 *         elif size == 0x05:             # <<<<<<<<<<<<<<
 *             write(b"\x85")
 *         elif size == 0x06:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":163
 *         elif size == 0x05:
 *             write(b"\x85")
 *         elif size == 0x06:             # <<<<<<<<<<<<<<
 *             write(b"\x86")
 *         elif size == 0x07:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x06) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":164
 *             write(b"\x85")
 *         elif size == 0x06:
 *             write(b"\x86")             # <<<<<<<<<<<<<<
 *         elif size == 0x07:
 *             write(b"\x87")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":163
 *         elif size == 0x05:
 *             write(b"\x85")
 *         elif size == 0x06:             # <<<<<<<<<<<<<<
 *             write(b"\x86")
 *         elif size == 0x07:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":165
 *         elif size == 0x06:
 *             write(b"\x86")
 *         elif size == 0x07:             # <<<<<<<<<<<<<<
 *             write(b"\x87")
 *         elif size == 0x08:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x07) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":166
 *             write(b"\x86")
 *         elif size == 0x07:
 *             write(b"\x87")             # <<<<<<<<<<<<<<
 *         elif size == 0x08:
 *             write(b"\x88")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":165
 *         elif size == 0x06:
 *             write(b"\x86")
 *         elif size == 0x07:             # <<<<<<<<<<<<<<
 *             write(b"\x87")
 *         elif size == 0x08:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":167
 *         elif size == 0x07:
 *             write(b"\x87")
 *         elif size == 0x08:             # <<<<<<<<<<<<<<
 *             write(b"\x88")
 *         elif size == 0x09:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x08) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":168
 *             write(b"\x87")
 *         elif size == 0x08:
 *             write(b"\x88")             # <<<<<<<<<<<<<<
 *         elif size == 0x09:
 *             write(b"\x89")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__44, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":167
 *         elif size == 0x07:
 *             write(b"\x87")
 *         elif size == 0x08:             # <<<<<<<<<<<<<<
 *             write(b"\x88")
 *         elif size == 0x09:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":169
 *         elif size == 0x08:
 *             write(b"\x88")
 *         elif size == 0x09:             # <<<<<<<<<<<<<<
 *             write(b"\x89")
 *         elif size == 0x0A:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x09) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":170
 *             write(b"\x88")
 *         elif size == 0x09:
 *             write(b"\x89")             # <<<<<<<<<<<<<<
 *         elif size == 0x0A:
 *             write(b"\x8A")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":169
 *         elif size == 0x08:
 *             write(b"\x88")
 *         elif size == 0x09:             # <<<<<<<<<<<<<<
 *             write(b"\x89")
 *         elif size == 0x0A:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":171
 *         elif size == 0x09:
 *             write(b"\x89")
 *         elif size == 0x0A:             # <<<<<<<<<<<<<<
 *             write(b"\x8A")
 *         elif size == 0x0B:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0A) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":172
 *             write(b"\x89")
 *         elif size == 0x0A:
 *             write(b"\x8A")             # <<<<<<<<<<<<<<
 *         elif size == 0x0B:
 *             write(b"\x8B")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":171
 *         elif size == 0x09:
 *             write(b"\x89")
 *         elif size == 0x0A:             # <<<<<<<<<<<<<<
 *             write(b"\x8A")
 *         elif size == 0x0B:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":173
 *         elif size == 0x0A:
 *             write(b"\x8A")
 *         elif size == 0x0B:             # <<<<<<<<<<<<<<
 *             write(b"\x8B")
 *         elif size == 0x0C:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0B) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":174
 *             write(b"\x8A")
 *         elif size == 0x0B:
 *             write(b"\x8B")             # <<<<<<<<<<<<<<
 *         elif size == 0x0C:
 *             write(b"\x8C")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__50, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":173
 *         elif size == 0x0A:
 *             write(b"\x8A")
 *         elif size == 0x0B:             # <<<<<<<<<<<<<<
 *             write(b"\x8B")
 *         elif size == 0x0C:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":175
 *         elif size == 0x0B:
 *             write(b"\x8B")
 *         elif size == 0x0C:             # <<<<<<<<<<<<<<
 *             write(b"\x8C")
 *         elif size == 0x0D:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0C) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":176
 *             write(b"\x8B")
 *         elif size == 0x0C:
 *             write(b"\x8C")             # <<<<<<<<<<<<<<
 *         elif size == 0x0D:
 *             write(b"\x8D")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__52, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":175
 *         elif size == 0x0B:
 *             write(b"\x8B")
 *         elif size == 0x0C:             # <<<<<<<<<<<<<<
 *             write(b"\x8C")
 *         elif size == 0x0D:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":177
 *         elif size == 0x0C:
 *             write(b"\x8C")
 *         elif size == 0x0D:             # <<<<<<<<<<<<<<
 *             write(b"\x8D")
 *         elif size == 0x0E:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0D) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":178
 *             write(b"\x8C")
 *         elif size == 0x0D:
 *             write(b"\x8D")             # <<<<<<<<<<<<<<
 *         elif size == 0x0E:
 *             write(b"\x8E")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__54, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":177
 *         elif size == 0x0C:
 *             write(b"\x8C")
 *         elif size == 0x0D:             # <<<<<<<<<<<<<<
 *             write(b"\x8D")
 *         elif size == 0x0E:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":179
 *         elif size == 0x0D:
 *             write(b"\x8D")
 *         elif size == 0x0E:             # <<<<<<<<<<<<<<
 *             write(b"\x8E")
 *         elif size == 0x0F:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0E) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":180
 *             write(b"\x8D")
 *         elif size == 0x0E:
 *             write(b"\x8E")             # <<<<<<<<<<<<<<
 *         elif size == 0x0F:
 *             write(b"\x8F")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":179
 *         elif size == 0x0D:
 *             write(b"\x8D")
 *         elif size == 0x0E:             # <<<<<<<<<<<<<<
 *             write(b"\x8E")
 *         elif size == 0x0F:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":181
 *         elif size == 0x0E:
 *             write(b"\x8E")
 *         elif size == 0x0F:             # <<<<<<<<<<<<<<
 *             write(b"\x8F")
 *         elif size < 0x100:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0F) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":182
 *             write(b"\x8E")
 *         elif size == 0x0F:
 *             write(b"\x8F")             # <<<<<<<<<<<<<<
 *         elif size < 0x100:
 *             write(b"\xD0")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":181
 *         elif size == 0x0E:
 *             write(b"\x8E")
 *         elif size == 0x0F:             # <<<<<<<<<<<<<<
 *             write(b"\x8F")
 *         elif size < 0x100:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":183
 *         elif size == 0x0F:
 *             write(b"\x8F")
 *         elif size < 0x100:             # <<<<<<<<<<<<<<
 *             write(b"\xD0")
 *             write(PACKED_UINT_8[size])
 */
  __pyx_t_2 = ((__pyx_v_size < 0x100) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":184
 *             write(b"\x8F")
 *         elif size < 0x100:
 *             write(b"\xD0")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__60, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":185
 *         elif size < 0x100:
 *             write(b"\xD0")
 *             write(PACKED_UINT_8[size])             # <<<<<<<<<<<<<<
 *         elif size < 0x10000:
 *             write(b"\xD1")
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_size, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_3 = __pyx_v_write; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":183
 *         elif size == 0x0F:
 *             write(b"\x8F")
 *         elif size < 0x100:             # <<<<<<<<<<<<<<
 *             write(b"\xD0")
 *             write(PACKED_UINT_8[size])
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":186
 *             write(b"\xD0")
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:             # <<<<<<<<<<<<<<
 *             write(b"\xD1")
 *             write(PACKED_UINT_16[size])
 */
  __pyx_t_2 = ((__pyx_v_size < 0x10000) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":187
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 *             write(b"\xD1")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__62, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":188
 *         elif size < 0x10000:
 *             write(b"\xD1")
 *             write(PACKED_UINT_16[size])             # <<<<<<<<<<<<<<
 *         elif size < 0x100000000:
 *             write(b"\xD2")
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_size, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_3 = __pyx_v_write; __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":186
 *             write(b"\xD0")
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:             # <<<<<<<<<<<<<<
 *             write(b"\xD1")
 *             write(PACKED_UINT_16[size])
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":189
 *             write(b"\xD1")
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:             # <<<<<<<<<<<<<<
 *             write(b"\xD2")
 *             write(struct_pack(">I", size))
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_int_4294967296, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":190
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 *             write(b"\xD2")             # <<<<<<<<<<<<<<
 *             write(struct_pack(">I", size))
 *         else:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__64, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "neo4j/packstream/_packer.pyx":191
 *         elif size < 0x100000000:
 *             write(b"\xD2")
 *             write(struct_pack(">I", size))             # <<<<<<<<<<<<<<
 *         else:
 *             raise OverflowError("String header size out of range")
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_struct_pack); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_s_I, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_s_I, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_I);
      __Pyx_GIVEREF(__pyx_kp_s_I);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_kp_s_I);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_5 = __pyx_v_write; __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (!__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_1};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_1};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8); __pyx_t_8 = NULL;
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_1);
        __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "neo4j/packstream/_packer.pyx":189
 *             write(b"\xD1")
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:             # <<<<<<<<<<<<<<
 *             write(b"\xD2")
 *             write(struct_pack(">I", size))
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":193
 *             write(struct_pack(">I", size))
 *         else:
 *             raise OverflowError("String header size out of range")             # <<<<<<<<<<<<<<
 * 
 *     cdef pack_list_header(self, int size):
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_OverflowError, __pyx_tuple__65, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 193, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "neo4j/packstream/_packer.pyx":149
 *             raise OverflowError("Bytes header size out of range")
 * 
 *     cdef pack_string_header(self, int size):             # <<<<<<<<<<<<<<
 *         write = self._write
 *         if size == 0x00:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_string_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_write);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "neo4j/packstream/_packer.pyx":195
 *             raise OverflowError("String header size out of range")
 * 
 *     cdef pack_list_header(self, int size):             # <<<<<<<<<<<<<<
 *         write = self._write
 *         if size == 0x00:
 */

static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_list_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_v_size) {
  PyObject *__pyx_v_write = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("pack_list_header", 0);

  /* "neo4j/packstream/_packer.pyx":196
 * 
 *     cdef pack_list_header(self, int size):
 *         write = self._write             # <<<<<<<<<<<<<<
 *         if size == 0x00:
 *             write(b"\x90")
 */
  __pyx_t_1 = __pyx_v_self->_write;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_write = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":197
 *     cdef pack_list_header(self, int size):
 *         write = self._write
 *         if size == 0x00:             # <<<<<<<<<<<<<<
 *             write(b"\x90")
 *         elif size == 0x01:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x00) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":198
 *         write = self._write
 *         if size == 0x00:
 *             write(b"\x90")             # <<<<<<<<<<<<<<
 *         elif size == 0x01:
 *             write(b"\x91")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__67, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":197
 *     cdef pack_list_header(self, int size):
 *         write = self._write
 *         if size == 0x00:             # <<<<<<<<<<<<<<
 *             write(b"\x90")
 *         elif size == 0x01:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":199
 *         if size == 0x00:
 *             write(b"\x90")
 *         elif size == 0x01:             # <<<<<<<<<<<<<<
 *             write(b"\x91")
 *         elif size == 0x02:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x01) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":200
 *             write(b"\x90")
 *         elif size == 0x01:
 *             write(b"\x91")             # <<<<<<<<<<<<<<
 *         elif size == 0x02:
 *             write(b"\x92")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__69, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":199
 *         if size == 0x00:
 *             write(b"\x90")
 *         elif size == 0x01:             # <<<<<<<<<<<<<<
 *             write(b"\x91")
 *         elif size == 0x02:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":201
 *         elif size == 0x01:
 *             write(b"\x91")
 *         elif size == 0x02:             # <<<<<<<<<<<<<<
 *             write(b"\x92")
 *         elif size == 0x03:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x02) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":202
 *             write(b"\x91")
 *         elif size == 0x02:
 *             write(b"\x92")             # <<<<<<<<<<<<<<
 *         elif size == 0x03:
 *             write(b"\x93")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__71, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":201
 *         elif size == 0x01:
 *             write(b"\x91")
 *         elif size == 0x02:             # <<<<<<<<<<<<<<
 *             write(b"\x92")
 *         elif size == 0x03:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":203
 *         elif size == 0x02:
 *             write(b"\x92")
 *         elif size == 0x03:             # <<<<<<<<<<<<<<
 *             write(b"\x93")
 *         elif size == 0x04:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x03) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":204
 *             write(b"\x92")
 *         elif size == 0x03:
 *             write(b"\x93")             # <<<<<<<<<<<<<<
 *         elif size == 0x04:
 *             write(b"\x94")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__73, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":203
 *         elif size == 0x02:
 *             write(b"\x92")
 *         elif size == 0x03:             # <<<<<<<<<<<<<<
 *             write(b"\x93")
 *         elif size == 0x04:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":205
 *         elif size == 0x03:
 *             write(b"\x93")
 *         elif size == 0x04:             # <<<<<<<<<<<<<<
 *             write(b"\x94")
 *         elif size == 0x05:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x04) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":206
 *             write(b"\x93")
 *         elif size == 0x04:
 *             write(b"\x94")             # <<<<<<<<<<<<<<
 *         elif size == 0x05:
 *             write(b"\x95")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__75, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":205
 *         elif size == 0x03:
 *             write(b"\x93")
 *         elif size == 0x04:             # <<<<<<<<<<<<<<
 *             write(b"\x94")
 *         elif size == 0x05:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":207
 *         elif size == 0x04:
 *             write(b"\x94")
 *         elif size == 0x05:             # <<<<<<<<<<<<<<
 *             write(b"\x95")
 *         elif size == 0x06:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x05) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":208
 *             write(b"\x94")
 *         elif size == 0x05:
 *             write(b"\x95")             # <<<<<<<<<<<<<<
 *         elif size == 0x06:
 *             write(b"\x96")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__77, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":207
 *         elif size == 0x04:
 *             write(b"\x94")
 *         elif size == 0x05:             # <<<<<<<<<<<<<<
 *             write(b"\x95")
 *         elif size == 0x06:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":209
 *         elif size == 0x05:
 *             write(b"\x95")
 *         elif size == 0x06:             # <<<<<<<<<<<<<<
 *             write(b"\x96")
 *         elif size == 0x07:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x06) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":210
 *             write(b"\x95")
 *         elif size == 0x06:
 *             write(b"\x96")             # <<<<<<<<<<<<<<
 *         elif size == 0x07:
 *             write(b"\x97")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__79, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":209
 *         elif size == 0x05:
 *             write(b"\x95")
 *         elif size == 0x06:             # <<<<<<<<<<<<<<
 *             write(b"\x96")
 *         elif size == 0x07:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":211
 *         elif size == 0x06:
 *             write(b"\x96")
 *         elif size == 0x07:             # <<<<<<<<<<<<<<
 *             write(b"\x97")
 *         elif size == 0x08:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x07) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":212
 *             write(b"\x96")
 *         elif size == 0x07:
 *             write(b"\x97")             # <<<<<<<<<<<<<<
 *         elif size == 0x08:
 *             write(b"\x98")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__81, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":211
 *         elif size == 0x06:
 *             write(b"\x96")
 *         elif size == 0x07:             # <<<<<<<<<<<<<<
 *             write(b"\x97")
 *         elif size == 0x08:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":213
 *         elif size == 0x07:
 *             write(b"\x97")
 *         elif size == 0x08:             # <<<<<<<<<<<<<<
 *             write(b"\x98")
 *         elif size == 0x09:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x08) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":214
 *             write(b"\x97")
 *         elif size == 0x08:
 *             write(b"\x98")             # <<<<<<<<<<<<<<
 *         elif size == 0x09:
 *             write(b"\x99")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__83, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":213
 *         elif size == 0x07:
 *             write(b"\x97")
 *         elif size == 0x08:             # <<<<<<<<<<<<<<
 *             write(b"\x98")
 *         elif size == 0x09:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":215
 *         elif size == 0x08:
 *             write(b"\x98")
 *         elif size == 0x09:             # <<<<<<<<<<<<<<
 *             write(b"\x99")
 *         elif size == 0x0A:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x09) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":216
 *             write(b"\x98")
 *         elif size == 0x09:
 *             write(b"\x99")             # <<<<<<<<<<<<<<
 *         elif size == 0x0A:
 *             write(b"\x9A")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__85, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":215
 *         elif size == 0x08:
 *             write(b"\x98")
 *         elif size == 0x09:             # <<<<<<<<<<<<<<
 *             write(b"\x99")
 *         elif size == 0x0A:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":217
 *         elif size == 0x09:
 *             write(b"\x99")
 *         elif size == 0x0A:             # <<<<<<<<<<<<<<
 *             write(b"\x9A")
 *         elif size == 0x0B:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0A) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":218
 *             write(b"\x99")
 *         elif size == 0x0A:
 *             write(b"\x9A")             # <<<<<<<<<<<<<<
 *         elif size == 0x0B:
 *             write(b"\x9B")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__87, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":217
 *         elif size == 0x09:
 *             write(b"\x99")
 *         elif size == 0x0A:             # <<<<<<<<<<<<<<
 *             write(b"\x9A")
 *         elif size == 0x0B:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":219
 *         elif size == 0x0A:
 *             write(b"\x9A")
 *         elif size == 0x0B:             # <<<<<<<<<<<<<<
 *             write(b"\x9B")
 *         elif size == 0x0C:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0B) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":220
 *             write(b"\x9A")
 *         elif size == 0x0B:
 *             write(b"\x9B")             # <<<<<<<<<<<<<<
 *         elif size == 0x0C:
 *             write(b"\x9C")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__89, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":219
 *         elif size == 0x0A:
 *             write(b"\x9A")
 *         elif size == 0x0B:             # <<<<<<<<<<<<<<
 *             write(b"\x9B")
 *         elif size == 0x0C:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":221
 *         elif size == 0x0B:
 *             write(b"\x9B")
 *         elif size == 0x0C:             # <<<<<<<<<<<<<<
 *             write(b"\x9C")
 *         elif size == 0x0D:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0C) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":222
 *             write(b"\x9B")
 *         elif size == 0x0C:
 *             write(b"\x9C")             # <<<<<<<<<<<<<<
 *         elif size == 0x0D:
 *             write(b"\x9D")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__91, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":221
 *         elif size == 0x0B:
 *             write(b"\x9B")
 *         elif size == 0x0C:             # <<<<<<<<<<<<<<
 *             write(b"\x9C")
 *         elif size == 0x0D:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":223
 *         elif size == 0x0C:
 *             write(b"\x9C")
 *         elif size == 0x0D:             # <<<<<<<<<<<<<<
 *             write(b"\x9D")
 *         elif size == 0x0E:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0D) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":224
 *             write(b"\x9C")
 *         elif size == 0x0D:
 *             write(b"\x9D")             # <<<<<<<<<<<<<<
 *         elif size == 0x0E:
 *             write(b"\x9E")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__93, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":223
 *         elif size == 0x0C:
 *             write(b"\x9C")
 *         elif size == 0x0D:             # <<<<<<<<<<<<<<
 *             write(b"\x9D")
 *         elif size == 0x0E:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":225
 *         elif size == 0x0D:
 *             write(b"\x9D")
 *         elif size == 0x0E:             # <<<<<<<<<<<<<<
 *             write(b"\x9E")
 *         elif size == 0x0F:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0E) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":226
 *             write(b"\x9D")
 *         elif size == 0x0E:
 *             write(b"\x9E")             # <<<<<<<<<<<<<<
 *         elif size == 0x0F:
 *             write(b"\x9F")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__95, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":225
 *         elif size == 0x0D:
 *             write(b"\x9D")
 *         elif size == 0x0E:             # <<<<<<<<<<<<<<
 *             write(b"\x9E")
 *         elif size == 0x0F:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":227
 *         elif size == 0x0E:
 *             write(b"\x9E")
 *         elif size == 0x0F:             # <<<<<<<<<<<<<<
 *             write(b"\x9F")
 *         elif size < 0x100:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0F) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":228
 *             write(b"\x9E")
 *         elif size == 0x0F:
 *             write(b"\x9F")             # <<<<<<<<<<<<<<
 *         elif size < 0x100:
 *             write(b"\xD4")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__97, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":227
 *         elif size == 0x0E:
 *             write(b"\x9E")
 *         elif size == 0x0F:             # <<<<<<<<<<<<<<
 *             write(b"\x9F")
 *         elif size < 0x100:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":229
 *         elif size == 0x0F:
 *             write(b"\x9F")
 *         elif size < 0x100:             # <<<<<<<<<<<<<<
 *             write(b"\xD4")
 *             write(PACKED_UINT_8[size])
 */
  __pyx_t_2 = ((__pyx_v_size < 0x100) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":230
 *             write(b"\x9F")
 *         elif size < 0x100:
 *             write(b"\xD4")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__99, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":231
 *         elif size < 0x100:
 *             write(b"\xD4")
 *             write(PACKED_UINT_8[size])             # <<<<<<<<<<<<<<
 *         elif size < 0x10000:
 *             write(b"\xD5")
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_size, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_3 = __pyx_v_write; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 231, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":229
 *         elif size == 0x0F:
 *             write(b"\x9F")
 *         elif size < 0x100:             # <<<<<<<<<<<<<<
 *             write(b"\xD4")
 *             write(PACKED_UINT_8[size])
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":232
 *             write(b"\xD4")
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:             # <<<<<<<<<<<<<<
 *             write(b"\xD5")
 *             write(PACKED_UINT_16[size])
 */
  __pyx_t_2 = ((__pyx_v_size < 0x10000) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":233
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 *             write(b"\xD5")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__101, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":234
 *         elif size < 0x10000:
 *             write(b"\xD5")
 *             write(PACKED_UINT_16[size])             # <<<<<<<<<<<<<<
 *         elif size < 0x100000000:
 *             write(b"\xD6")
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_size, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_3 = __pyx_v_write; __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 234, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":232
 *             write(b"\xD4")
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:             # <<<<<<<<<<<<<<
 *             write(b"\xD5")
 *             write(PACKED_UINT_16[size])
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":235
 *             write(b"\xD5")
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:             # <<<<<<<<<<<<<<
 *             write(b"\xD6")
 *             write(struct_pack(">I", size))
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_int_4294967296, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":236
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 *             write(b"\xD6")             # <<<<<<<<<<<<<<
 *             write(struct_pack(">I", size))
 *         else:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__103, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "neo4j/packstream/_packer.pyx":237
 *         elif size < 0x100000000:
 *             write(b"\xD6")
 *             write(struct_pack(">I", size))             # <<<<<<<<<<<<<<
 *         else:
 *             raise OverflowError("List header size out of range")
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_struct_pack); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_s_I, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_s_I, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_I);
      __Pyx_GIVEREF(__pyx_kp_s_I);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_kp_s_I);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_5 = __pyx_v_write; __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (!__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_1};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 237, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_1};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 237, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8); __pyx_t_8 = NULL;
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_1);
        __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "neo4j/packstream/_packer.pyx":235
 *             write(b"\xD5")
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:             # <<<<<<<<<<<<<<
 *             write(b"\xD6")
 *             write(struct_pack(">I", size))
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":239
 *             write(struct_pack(">I", size))
 *         else:
 *             raise OverflowError("List header size out of range")             # <<<<<<<<<<<<<<
 * 
 *     cpdef pack_list_stream_header(self):
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_OverflowError, __pyx_tuple__104, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 239, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "neo4j/packstream/_packer.pyx":195
 *             raise OverflowError("String header size out of range")
 * 
 *     cdef pack_list_header(self, int size):             # <<<<<<<<<<<<<<
 *         write = self._write
 *         if size == 0x00:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_list_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_write);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "neo4j/packstream/_packer.pyx":241
 *             raise OverflowError("List header size out of range")
 * 
 *     cpdef pack_list_stream_header(self):             # <<<<<<<<<<<<<<
 *         self._write(b"\xD7")
 * 
 */

static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_5pack_list_stream_header(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_list_stream_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("pack_list_stream_header", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pack_list_stream_header); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5neo4j_10packstream_7_packer_6Packer_5pack_list_stream_header)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 241, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 241, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "neo4j/packstream/_packer.pyx":242
 * 
 *     cpdef pack_list_stream_header(self):
 *         self._write(b"\xD7")             # <<<<<<<<<<<<<<
 * 
 *     cdef pack_map_header(self, int size):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_self->_write, __pyx_tuple__106, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":241
 *             raise OverflowError("List header size out of range")
 * 
 *     cpdef pack_list_stream_header(self):             # <<<<<<<<<<<<<<
 *         self._write(b"\xD7")
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_list_stream_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_5pack_list_stream_header(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_5pack_list_stream_header(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_list_stream_header (wrapper)", 0);
  __pyx_r = __pyx_pf_5neo4j_10packstream_7_packer_6Packer_4pack_list_stream_header(((struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5neo4j_10packstream_7_packer_6Packer_4pack_list_stream_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pack_list_stream_header", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_list_stream_header(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_list_stream_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "neo4j/packstream/_packer.pyx":244
 *         self._write(b"\xD7")
 * 
 *     cdef pack_map_header(self, int size):             # <<<<<<<<<<<<<<
 *         write = self._write
 *         if size == 0x00:
 */

static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_map_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_v_size) {
  PyObject *__pyx_v_write = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("pack_map_header", 0);

  /* "neo4j/packstream/_packer.pyx":245
 * 
 *     cdef pack_map_header(self, int size):
 *         write = self._write             # <<<<<<<<<<<<<<
 *         if size == 0x00:
 *             write(b"\xA0")
 */
  __pyx_t_1 = __pyx_v_self->_write;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_write = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":246
 *     cdef pack_map_header(self, int size):
 *         write = self._write
 *         if size == 0x00:             # <<<<<<<<<<<<<<
 *             write(b"\xA0")
 *         elif size == 0x01:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x00) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":247
 *         write = self._write
 *         if size == 0x00:
 *             write(b"\xA0")             # <<<<<<<<<<<<<<
 *         elif size == 0x01:
 *             write(b"\xA1")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__108, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":246
 *     cdef pack_map_header(self, int size):
 *         write = self._write
 *         if size == 0x00:             # <<<<<<<<<<<<<<
 *             write(b"\xA0")
 *         elif size == 0x01:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":248
 *         if size == 0x00:
 *             write(b"\xA0")
 *         elif size == 0x01:             # <<<<<<<<<<<<<<
 *             write(b"\xA1")
 *         elif size == 0x02:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x01) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":249
 *             write(b"\xA0")
 *         elif size == 0x01:
 *             write(b"\xA1")             # <<<<<<<<<<<<<<
 *         elif size == 0x02:
 *             write(b"\xA2")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__110, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":248
 *         if size == 0x00:
 *             write(b"\xA0")
 *         elif size == 0x01:             # <<<<<<<<<<<<<<
 *             write(b"\xA1")
 *         elif size == 0x02:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":250
 *         elif size == 0x01:
 *             write(b"\xA1")
 *         elif size == 0x02:             # <<<<<<<<<<<<<<
 *             write(b"\xA2")
 *         elif size == 0x03:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x02) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":251
 *             write(b"\xA1")
 *         elif size == 0x02:
 *             write(b"\xA2")             # <<<<<<<<<<<<<<
 *         elif size == 0x03:
 *             write(b"\xA3")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__112, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":250
 *         elif size == 0x01:
 *             write(b"\xA1")
 *         elif size == 0x02:             # <<<<<<<<<<<<<<
 *             write(b"\xA2")
 *         elif size == 0x03:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":252
 *         elif size == 0x02:
 *             write(b"\xA2")
 *         elif size == 0x03:             # <<<<<<<<<<<<<<
 *             write(b"\xA3")
 *         elif size == 0x04:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x03) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":253
 *             write(b"\xA2")
 *         elif size == 0x03:
 *             write(b"\xA3")             # <<<<<<<<<<<<<<
 *         elif size == 0x04:
 *             write(b"\xA4")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__114, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":252
 *         elif size == 0x02:
 *             write(b"\xA2")
 *         elif size == 0x03:             # <<<<<<<<<<<<<<
 *             write(b"\xA3")
 *         elif size == 0x04:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":254
 *         elif size == 0x03:
 *             write(b"\xA3")
 *         elif size == 0x04:             # <<<<<<<<<<<<<<
 *             write(b"\xA4")
 *         elif size == 0x05:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x04) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":255
 *             write(b"\xA3")
 *         elif size == 0x04:
 *             write(b"\xA4")             # <<<<<<<<<<<<<<
 *         elif size == 0x05:
 *             write(b"\xA5")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__116, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":254
 *         elif size == 0x03:
 *             write(b"\xA3")
 *         elif size == 0x04:             # <<<<<<<<<<<<<<
 *             write(b"\xA4")
 *         elif size == 0x05:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":256
 *         elif size == 0x04:
 *             write(b"\xA4")
 *         elif size == 0x05:             # <<<<<<<<<<<<<<
 *             write(b"\xA5")
 *         elif size == 0x06:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x05) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":257
 *             write(b"\xA4")
 *         elif size == 0x05:
 *             write(b"\xA5")             # <<<<<<<<<<<<<<
 *         elif size == 0x06:
 *             write(b"\xA6")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__118, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":256
 *         elif size == 0x04:
 *             write(b"\xA4")
 *         elif size == 0x05:             # <<<<<<<<<<<<<<
 *             write(b"\xA5")
 *         elif size == 0x06:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":258
 *         elif size == 0x05:
 *             write(b"\xA5")
 *         elif size == 0x06:             # <<<<<<<<<<<<<<
 *             write(b"\xA6")
 *         elif size == 0x07:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x06) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":259
 *             write(b"\xA5")
 *         elif size == 0x06:
 *             write(b"\xA6")             # <<<<<<<<<<<<<<
 *         elif size == 0x07:
 *             write(b"\xA7")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__120, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":258
 *         elif size == 0x05:
 *             write(b"\xA5")
 *         elif size == 0x06:             # <<<<<<<<<<<<<<
 *             write(b"\xA6")
 *         elif size == 0x07:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":260
 *         elif size == 0x06:
 *             write(b"\xA6")
 *         elif size == 0x07:             # <<<<<<<<<<<<<<
 *             write(b"\xA7")
 *         elif size == 0x08:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x07) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":261
 *             write(b"\xA6")
 *         elif size == 0x07:
 *             write(b"\xA7")             # <<<<<<<<<<<<<<
 *         elif size == 0x08:
 *             write(b"\xA8")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__122, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":260
 *         elif size == 0x06:
 *             write(b"\xA6")
 *         elif size == 0x07:             # <<<<<<<<<<<<<<
 *             write(b"\xA7")
 *         elif size == 0x08:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":262
 *         elif size == 0x07:
 *             write(b"\xA7")
 *         elif size == 0x08:             # <<<<<<<<<<<<<<
 *             write(b"\xA8")
 *         elif size == 0x09:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x08) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":263
 *             write(b"\xA7")
 *         elif size == 0x08:
 *             write(b"\xA8")             # <<<<<<<<<<<<<<
 *         elif size == 0x09:
 *             write(b"\xA9")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__124, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":262
 *         elif size == 0x07:
 *             write(b"\xA7")
 *         elif size == 0x08:             # <<<<<<<<<<<<<<
 *             write(b"\xA8")
 *         elif size == 0x09:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":264
 *         elif size == 0x08:
 *             write(b"\xA8")
 *         elif size == 0x09:             # <<<<<<<<<<<<<<
 *             write(b"\xA9")
 *         elif size == 0x0A:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x09) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":265
 *             write(b"\xA8")
 *         elif size == 0x09:
 *             write(b"\xA9")             # <<<<<<<<<<<<<<
 *         elif size == 0x0A:
 *             write(b"\xAA")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__126, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":264
 *         elif size == 0x08:
 *             write(b"\xA8")
 *         elif size == 0x09:             # <<<<<<<<<<<<<<
 *             write(b"\xA9")
 *         elif size == 0x0A:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":266
 *         elif size == 0x09:
 *             write(b"\xA9")
 *         elif size == 0x0A:             # <<<<<<<<<<<<<<
 *             write(b"\xAA")
 *         elif size == 0x0B:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0A) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":267
 *             write(b"\xA9")
 *         elif size == 0x0A:
 *             write(b"\xAA")             # <<<<<<<<<<<<<<
 *         elif size == 0x0B:
 *             write(b"\xAB")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__128, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":266
 *         elif size == 0x09:
 *             write(b"\xA9")
 *         elif size == 0x0A:             # <<<<<<<<<<<<<<
 *             write(b"\xAA")
 *         elif size == 0x0B:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":268
 *         elif size == 0x0A:
 *             write(b"\xAA")
 *         elif size == 0x0B:             # <<<<<<<<<<<<<<
 *             write(b"\xAB")
 *         elif size == 0x0C:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0B) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":269
 *             write(b"\xAA")
 *         elif size == 0x0B:
 *             write(b"\xAB")             # <<<<<<<<<<<<<<
 *         elif size == 0x0C:
 *             write(b"\xAC")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__130, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":268
 *         elif size == 0x0A:
 *             write(b"\xAA")
 *         elif size == 0x0B:             # <<<<<<<<<<<<<<
 *             write(b"\xAB")
 *         elif size == 0x0C:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":270
 *         elif size == 0x0B:
 *             write(b"\xAB")
 *         elif size == 0x0C:             # <<<<<<<<<<<<<<
 *             write(b"\xAC")
 *         elif size == 0x0D:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0C) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":271
 *             write(b"\xAB")
 *         elif size == 0x0C:
 *             write(b"\xAC")             # <<<<<<<<<<<<<<
 *         elif size == 0x0D:
 *             write(b"\xAD")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__132, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":270
 *         elif size == 0x0B:
 *             write(b"\xAB")
 *         elif size == 0x0C:             # <<<<<<<<<<<<<<
 *             write(b"\xAC")
 *         elif size == 0x0D:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":272
 *         elif size == 0x0C:
 *             write(b"\xAC")
 *         elif size == 0x0D:             # <<<<<<<<<<<<<<
 *             write(b"\xAD")
 *         elif size == 0x0E:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0D) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":273
 *             write(b"\xAC")
 *         elif size == 0x0D:
 *             write(b"\xAD")             # <<<<<<<<<<<<<<
 *         elif size == 0x0E:
 *             write(b"\xAE")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__134, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":272
 *         elif size == 0x0C:
 *             write(b"\xAC")
 *         elif size == 0x0D:             # <<<<<<<<<<<<<<
 *             write(b"\xAD")
 *         elif size == 0x0E:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":274
 *         elif size == 0x0D:
 *             write(b"\xAD")
 *         elif size == 0x0E:             # <<<<<<<<<<<<<<
 *             write(b"\xAE")
 *         elif size == 0x0F:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0E) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":275
 *             write(b"\xAD")
 *         elif size == 0x0E:
 *             write(b"\xAE")             # <<<<<<<<<<<<<<
 *         elif size == 0x0F:
 *             write(b"\xAF")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__136, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":274
 *         elif size == 0x0D:
 *             write(b"\xAD")
 *         elif size == 0x0E:             # <<<<<<<<<<<<<<
 *             write(b"\xAE")
 *         elif size == 0x0F:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":276
 *         elif size == 0x0E:
 *             write(b"\xAE")
 *         elif size == 0x0F:             # <<<<<<<<<<<<<<
 *             write(b"\xAF")
 *         elif size < 0x100:
 */
  __pyx_t_2 = ((__pyx_v_size == 0x0F) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":277
 *             write(b"\xAE")
 *         elif size == 0x0F:
 *             write(b"\xAF")             # <<<<<<<<<<<<<<
 *         elif size < 0x100:
 *             write(b"\xD8")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__138, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":276
 *         elif size == 0x0E:
 *             write(b"\xAE")
 *         elif size == 0x0F:             # <<<<<<<<<<<<<<
 *             write(b"\xAF")
 *         elif size < 0x100:
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":278
 *         elif size == 0x0F:
 *             write(b"\xAF")
 *         elif size < 0x100:             # <<<<<<<<<<<<<<
 *             write(b"\xD8")
 *             write(PACKED_UINT_8[size])
 */
  __pyx_t_2 = ((__pyx_v_size < 0x100) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":279
 *             write(b"\xAF")
 *         elif size < 0x100:
 *             write(b"\xD8")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__140, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":280
 *         elif size < 0x100:
 *             write(b"\xD8")
 *             write(PACKED_UINT_8[size])             # <<<<<<<<<<<<<<
 *         elif size < 0x10000:
 *             write(b"\xD9")
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_size, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_3 = __pyx_v_write; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":278
 *         elif size == 0x0F:
 *             write(b"\xAF")
 *         elif size < 0x100:             # <<<<<<<<<<<<<<
 *             write(b"\xD8")
 *             write(PACKED_UINT_8[size])
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":281
 *             write(b"\xD8")
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:             # <<<<<<<<<<<<<<
 *             write(b"\xD9")
 *             write(PACKED_UINT_16[size])
 */
  __pyx_t_2 = ((__pyx_v_size < 0x10000) != 0);
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":282
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 *             write(b"\xD9")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__142, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":283
 *         elif size < 0x10000:
 *             write(b"\xD9")
 *             write(PACKED_UINT_16[size])             # <<<<<<<<<<<<<<
 *         elif size < 0x100000000:
 *             write(b"\xDA")
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_size, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_3 = __pyx_v_write; __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 283, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":281
 *             write(b"\xD8")
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:             # <<<<<<<<<<<<<<
 *             write(b"\xD9")
 *             write(PACKED_UINT_16[size])
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":284
 *             write(b"\xD9")
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:             # <<<<<<<<<<<<<<
 *             write(b"\xDA")
 *             write(struct_pack(">I", size))
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_int_4294967296, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "neo4j/packstream/_packer.pyx":285
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 *             write(b"\xDA")             # <<<<<<<<<<<<<<
 *             write(struct_pack(">I", size))
 *         else:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__144, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "neo4j/packstream/_packer.pyx":286
 *         elif size < 0x100000000:
 *             write(b"\xDA")
 *             write(struct_pack(">I", size))             # <<<<<<<<<<<<<<
 *         else:
 *             raise OverflowError("Map header size out of range")
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_struct_pack); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_s_I, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_s_I, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_I);
      __Pyx_GIVEREF(__pyx_kp_s_I);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_kp_s_I);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_5 = __pyx_v_write; __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (!__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_1};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_1};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 286, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8); __pyx_t_8 = NULL;
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_1);
        __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "neo4j/packstream/_packer.pyx":284
 *             write(b"\xD9")
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:             # <<<<<<<<<<<<<<
 *             write(b"\xDA")
 *             write(struct_pack(">I", size))
 */
    goto __pyx_L3;
  }

  /* "neo4j/packstream/_packer.pyx":288
 *             write(struct_pack(">I", size))
 *         else:
 *             raise OverflowError("Map header size out of range")             # <<<<<<<<<<<<<<
 * 
 *     cpdef pack_map_stream_header(self):
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_OverflowError, __pyx_tuple__145, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 288, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "neo4j/packstream/_packer.pyx":244
 *         self._write(b"\xD7")
 * 
 *     cdef pack_map_header(self, int size):             # <<<<<<<<<<<<<<
 *         write = self._write
 *         if size == 0x00:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_map_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_write);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "neo4j/packstream/_packer.pyx":290
 *             raise OverflowError("Map header size out of range")
 * 
 *     cpdef pack_map_stream_header(self):             # <<<<<<<<<<<<<<
 *         self._write(b"\xDB")
 * 
 */

static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_7pack_map_stream_header(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_map_stream_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("pack_map_stream_header", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pack_map_stream_header); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5neo4j_10packstream_7_packer_6Packer_7pack_map_stream_header)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 290, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 290, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "neo4j/packstream/_packer.pyx":291
 * 
 *     cpdef pack_map_stream_header(self):
 *         self._write(b"\xDB")             # <<<<<<<<<<<<<<
 * 
 *     cpdef pack_struct(self, bytes signature, tuple fields):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_self->_write, __pyx_tuple__147, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":290
 *             raise OverflowError("Map header size out of range")
 * 
 *     cpdef pack_map_stream_header(self):             # <<<<<<<<<<<<<<
 *         self._write(b"\xDB")
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_map_stream_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_7pack_map_stream_header(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_7pack_map_stream_header(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_map_stream_header (wrapper)", 0);
  __pyx_r = __pyx_pf_5neo4j_10packstream_7_packer_6Packer_6pack_map_stream_header(((struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5neo4j_10packstream_7_packer_6Packer_6pack_map_stream_header(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pack_map_stream_header", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_map_stream_header(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_map_stream_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "neo4j/packstream/_packer.pyx":293
 *         self._write(b"\xDB")
 * 
 *     cpdef pack_struct(self, bytes signature, tuple fields):             # <<<<<<<<<<<<<<
 *         if len(signature) != 1:
 *             raise ValueError("Structure signature must be a single byte value")
 */

static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_9pack_struct(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_struct(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_signature, PyObject *__pyx_v_fields, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_write = NULL;
  Py_ssize_t __pyx_v_size;
  PyObject *__pyx_v_field = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("pack_struct", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pack_struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5neo4j_10packstream_7_packer_6Packer_9pack_struct)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_signature, __pyx_v_fields};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_signature, __pyx_v_fields};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 293, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(__pyx_v_signature);
        __Pyx_GIVEREF(__pyx_v_signature);
        PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_signature);
        __Pyx_INCREF(__pyx_v_fields);
        __Pyx_GIVEREF(__pyx_v_fields);
        PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_fields);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "neo4j/packstream/_packer.pyx":294
 * 
 *     cpdef pack_struct(self, bytes signature, tuple fields):
 *         if len(signature) != 1:             # <<<<<<<<<<<<<<
 *             raise ValueError("Structure signature must be a single byte value")
 *         write = self._write
 */
  if (unlikely(__pyx_v_signature == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 294, __pyx_L1_error)
  }
  __pyx_t_7 = PyBytes_GET_SIZE(__pyx_v_signature); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 294, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 != 1) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":295
 *     cpdef pack_struct(self, bytes signature, tuple fields):
 *         if len(signature) != 1:
 *             raise ValueError("Structure signature must be a single byte value")             # <<<<<<<<<<<<<<
 *         write = self._write
 *         size = len(fields)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__148, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 295, __pyx_L1_error)

    /* "neo4j/packstream/_packer.pyx":294
 * 
 *     cpdef pack_struct(self, bytes signature, tuple fields):
 *         if len(signature) != 1:             # <<<<<<<<<<<<<<
 *             raise ValueError("Structure signature must be a single byte value")
 *         write = self._write
 */
  }

  /* "neo4j/packstream/_packer.pyx":296
 *         if len(signature) != 1:
 *             raise ValueError("Structure signature must be a single byte value")
 *         write = self._write             # <<<<<<<<<<<<<<
 *         size = len(fields)
 *         if size == 0x00:
 */
  __pyx_t_1 = __pyx_v_self->_write;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_write = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":297
 *             raise ValueError("Structure signature must be a single byte value")
 *         write = self._write
 *         size = len(fields)             # <<<<<<<<<<<<<<
 *         if size == 0x00:
 *             write(b"\xB0")
 */
  if (unlikely(__pyx_v_fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 297, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v_fields); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 297, __pyx_L1_error)
  __pyx_v_size = __pyx_t_7;

  /* "neo4j/packstream/_packer.pyx":298
 *         write = self._write
 *         size = len(fields)
 *         if size == 0x00:             # <<<<<<<<<<<<<<
 *             write(b"\xB0")
 *         elif size == 0x01:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x00) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":299
 *         size = len(fields)
 *         if size == 0x00:
 *             write(b"\xB0")             # <<<<<<<<<<<<<<
 *         elif size == 0x01:
 *             write(b"\xB1")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__150, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":298
 *         write = self._write
 *         size = len(fields)
 *         if size == 0x00:             # <<<<<<<<<<<<<<
 *             write(b"\xB0")
 *         elif size == 0x01:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":300
 *         if size == 0x00:
 *             write(b"\xB0")
 *         elif size == 0x01:             # <<<<<<<<<<<<<<
 *             write(b"\xB1")
 *         elif size == 0x02:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x01) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":301
 *             write(b"\xB0")
 *         elif size == 0x01:
 *             write(b"\xB1")             # <<<<<<<<<<<<<<
 *         elif size == 0x02:
 *             write(b"\xB2")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__152, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":300
 *         if size == 0x00:
 *             write(b"\xB0")
 *         elif size == 0x01:             # <<<<<<<<<<<<<<
 *             write(b"\xB1")
 *         elif size == 0x02:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":302
 *         elif size == 0x01:
 *             write(b"\xB1")
 *         elif size == 0x02:             # <<<<<<<<<<<<<<
 *             write(b"\xB2")
 *         elif size == 0x03:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x02) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":303
 *             write(b"\xB1")
 *         elif size == 0x02:
 *             write(b"\xB2")             # <<<<<<<<<<<<<<
 *         elif size == 0x03:
 *             write(b"\xB3")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__154, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":302
 *         elif size == 0x01:
 *             write(b"\xB1")
 *         elif size == 0x02:             # <<<<<<<<<<<<<<
 *             write(b"\xB2")
 *         elif size == 0x03:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":304
 *         elif size == 0x02:
 *             write(b"\xB2")
 *         elif size == 0x03:             # <<<<<<<<<<<<<<
 *             write(b"\xB3")
 *         elif size == 0x04:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x03) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":305
 *             write(b"\xB2")
 *         elif size == 0x03:
 *             write(b"\xB3")             # <<<<<<<<<<<<<<
 *         elif size == 0x04:
 *             write(b"\xB4")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__156, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":304
 *         elif size == 0x02:
 *             write(b"\xB2")
 *         elif size == 0x03:             # <<<<<<<<<<<<<<
 *             write(b"\xB3")
 *         elif size == 0x04:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":306
 *         elif size == 0x03:
 *             write(b"\xB3")
 *         elif size == 0x04:             # <<<<<<<<<<<<<<
 *             write(b"\xB4")
 *         elif size == 0x05:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x04) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":307
 *             write(b"\xB3")
 *         elif size == 0x04:
 *             write(b"\xB4")             # <<<<<<<<<<<<<<
 *         elif size == 0x05:
 *             write(b"\xB5")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__158, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":306
 *         elif size == 0x03:
 *             write(b"\xB3")
 *         elif size == 0x04:             # <<<<<<<<<<<<<<
 *             write(b"\xB4")
 *         elif size == 0x05:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":308
 *         elif size == 0x04:
 *             write(b"\xB4")
 *         elif size == 0x05:             # <<<<<<<<<<<<<<
 *             write(b"\xB5")
 *         elif size == 0x06:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x05) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":309
 *             write(b"\xB4")
 *         elif size == 0x05:
 *             write(b"\xB5")             # <<<<<<<<<<<<<<
 *         elif size == 0x06:
 *             write(b"\xB6")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__160, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":308
 *         elif size == 0x04:
 *             write(b"\xB4")
 *         elif size == 0x05:             # <<<<<<<<<<<<<<
 *             write(b"\xB5")
 *         elif size == 0x06:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":310
 *         elif size == 0x05:
 *             write(b"\xB5")
 *         elif size == 0x06:             # <<<<<<<<<<<<<<
 *             write(b"\xB6")
 *         elif size == 0x07:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x06) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":311
 *             write(b"\xB5")
 *         elif size == 0x06:
 *             write(b"\xB6")             # <<<<<<<<<<<<<<
 *         elif size == 0x07:
 *             write(b"\xB7")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__162, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 311, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":310
 *         elif size == 0x05:
 *             write(b"\xB5")
 *         elif size == 0x06:             # <<<<<<<<<<<<<<
 *             write(b"\xB6")
 *         elif size == 0x07:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":312
 *         elif size == 0x06:
 *             write(b"\xB6")
 *         elif size == 0x07:             # <<<<<<<<<<<<<<
 *             write(b"\xB7")
 *         elif size == 0x08:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x07) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":313
 *             write(b"\xB6")
 *         elif size == 0x07:
 *             write(b"\xB7")             # <<<<<<<<<<<<<<
 *         elif size == 0x08:
 *             write(b"\xB8")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__164, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":312
 *         elif size == 0x06:
 *             write(b"\xB6")
 *         elif size == 0x07:             # <<<<<<<<<<<<<<
 *             write(b"\xB7")
 *         elif size == 0x08:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":314
 *         elif size == 0x07:
 *             write(b"\xB7")
 *         elif size == 0x08:             # <<<<<<<<<<<<<<
 *             write(b"\xB8")
 *         elif size == 0x09:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x08) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":315
 *             write(b"\xB7")
 *         elif size == 0x08:
 *             write(b"\xB8")             # <<<<<<<<<<<<<<
 *         elif size == 0x09:
 *             write(b"\xB9")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__166, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":314
 *         elif size == 0x07:
 *             write(b"\xB7")
 *         elif size == 0x08:             # <<<<<<<<<<<<<<
 *             write(b"\xB8")
 *         elif size == 0x09:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":316
 *         elif size == 0x08:
 *             write(b"\xB8")
 *         elif size == 0x09:             # <<<<<<<<<<<<<<
 *             write(b"\xB9")
 *         elif size == 0x0A:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x09) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":317
 *             write(b"\xB8")
 *         elif size == 0x09:
 *             write(b"\xB9")             # <<<<<<<<<<<<<<
 *         elif size == 0x0A:
 *             write(b"\xBA")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__168, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":316
 *         elif size == 0x08:
 *             write(b"\xB8")
 *         elif size == 0x09:             # <<<<<<<<<<<<<<
 *             write(b"\xB9")
 *         elif size == 0x0A:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":318
 *         elif size == 0x09:
 *             write(b"\xB9")
 *         elif size == 0x0A:             # <<<<<<<<<<<<<<
 *             write(b"\xBA")
 *         elif size == 0x0B:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x0A) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":319
 *             write(b"\xB9")
 *         elif size == 0x0A:
 *             write(b"\xBA")             # <<<<<<<<<<<<<<
 *         elif size == 0x0B:
 *             write(b"\xBB")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__170, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":318
 *         elif size == 0x09:
 *             write(b"\xB9")
 *         elif size == 0x0A:             # <<<<<<<<<<<<<<
 *             write(b"\xBA")
 *         elif size == 0x0B:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":320
 *         elif size == 0x0A:
 *             write(b"\xBA")
 *         elif size == 0x0B:             # <<<<<<<<<<<<<<
 *             write(b"\xBB")
 *         elif size == 0x0C:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x0B) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":321
 *             write(b"\xBA")
 *         elif size == 0x0B:
 *             write(b"\xBB")             # <<<<<<<<<<<<<<
 *         elif size == 0x0C:
 *             write(b"\xBC")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__172, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":320
 *         elif size == 0x0A:
 *             write(b"\xBA")
 *         elif size == 0x0B:             # <<<<<<<<<<<<<<
 *             write(b"\xBB")
 *         elif size == 0x0C:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":322
 *         elif size == 0x0B:
 *             write(b"\xBB")
 *         elif size == 0x0C:             # <<<<<<<<<<<<<<
 *             write(b"\xBC")
 *         elif size == 0x0D:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x0C) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":323
 *             write(b"\xBB")
 *         elif size == 0x0C:
 *             write(b"\xBC")             # <<<<<<<<<<<<<<
 *         elif size == 0x0D:
 *             write(b"\xBD")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__174, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":322
 *         elif size == 0x0B:
 *             write(b"\xBB")
 *         elif size == 0x0C:             # <<<<<<<<<<<<<<
 *             write(b"\xBC")
 *         elif size == 0x0D:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":324
 *         elif size == 0x0C:
 *             write(b"\xBC")
 *         elif size == 0x0D:             # <<<<<<<<<<<<<<
 *             write(b"\xBD")
 *         elif size == 0x0E:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x0D) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":325
 *             write(b"\xBC")
 *         elif size == 0x0D:
 *             write(b"\xBD")             # <<<<<<<<<<<<<<
 *         elif size == 0x0E:
 *             write(b"\xBE")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__176, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":324
 *         elif size == 0x0C:
 *             write(b"\xBC")
 *         elif size == 0x0D:             # <<<<<<<<<<<<<<
 *             write(b"\xBD")
 *         elif size == 0x0E:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":326
 *         elif size == 0x0D:
 *             write(b"\xBD")
 *         elif size == 0x0E:             # <<<<<<<<<<<<<<
 *             write(b"\xBE")
 *         elif size == 0x0F:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x0E) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":327
 *             write(b"\xBD")
 *         elif size == 0x0E:
 *             write(b"\xBE")             # <<<<<<<<<<<<<<
 *         elif size == 0x0F:
 *             write(b"\xBF")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__178, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":326
 *         elif size == 0x0D:
 *             write(b"\xBD")
 *         elif size == 0x0E:             # <<<<<<<<<<<<<<
 *             write(b"\xBE")
 *         elif size == 0x0F:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":328
 *         elif size == 0x0E:
 *             write(b"\xBE")
 *         elif size == 0x0F:             # <<<<<<<<<<<<<<
 *             write(b"\xBF")
 *         elif size < 0x100:
 */
  __pyx_t_8 = ((__pyx_v_size == 0x0F) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":329
 *             write(b"\xBE")
 *         elif size == 0x0F:
 *             write(b"\xBF")             # <<<<<<<<<<<<<<
 *         elif size < 0x100:
 *             write(b"\xDC")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__180, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":328
 *         elif size == 0x0E:
 *             write(b"\xBE")
 *         elif size == 0x0F:             # <<<<<<<<<<<<<<
 *             write(b"\xBF")
 *         elif size < 0x100:
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":330
 *         elif size == 0x0F:
 *             write(b"\xBF")
 *         elif size < 0x100:             # <<<<<<<<<<<<<<
 *             write(b"\xDC")
 *             write(PACKED_UINT_8[size])
 */
  __pyx_t_8 = ((__pyx_v_size < 0x100) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":331
 *             write(b"\xBF")
 *         elif size < 0x100:
 *             write(b"\xDC")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__182, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":332
 *         elif size < 0x100:
 *             write(b"\xDC")
 *             write(PACKED_UINT_8[size])             # <<<<<<<<<<<<<<
 *         elif size < 0x10000:
 *             write(b"\xDD")
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_size, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_2 = __pyx_v_write; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_t_3};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_t_3};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      {
        __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 332, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":330
 *         elif size == 0x0F:
 *             write(b"\xBF")
 *         elif size < 0x100:             # <<<<<<<<<<<<<<
 *             write(b"\xDC")
 *             write(PACKED_UINT_8[size])
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":333
 *             write(b"\xDC")
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:             # <<<<<<<<<<<<<<
 *             write(b"\xDD")
 *             write(PACKED_UINT_16[size])
 */
  __pyx_t_8 = ((__pyx_v_size < 0x10000) != 0);
  if (__pyx_t_8) {

    /* "neo4j/packstream/_packer.pyx":334
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 *             write(b"\xDD")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_16[size])
 *         else:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_write, __pyx_tuple__184, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":335
 *         elif size < 0x10000:
 *             write(b"\xDD")
 *             write(PACKED_UINT_16[size])             # <<<<<<<<<<<<<<
 *         else:
 *             raise OverflowError("Structure size out of range")
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_PACKED_UINT_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_size, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v_write);
    __pyx_t_2 = __pyx_v_write; __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (!__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "neo4j/packstream/_packer.pyx":333
 *             write(b"\xDC")
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:             # <<<<<<<<<<<<<<
 *             write(b"\xDD")
 *             write(PACKED_UINT_16[size])
 */
    goto __pyx_L4;
  }

  /* "neo4j/packstream/_packer.pyx":337
 *             write(PACKED_UINT_16[size])
 *         else:
 *             raise OverflowError("Structure size out of range")             # <<<<<<<<<<<<<<
 *         write(signature)
 *         for field in fields:
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_OverflowError, __pyx_tuple__185, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 337, __pyx_L1_error)
  }
  __pyx_L4:;

  /* "neo4j/packstream/_packer.pyx":338
 *         else:
 *             raise OverflowError("Structure size out of range")
 *         write(signature)             # <<<<<<<<<<<<<<
 *         for field in fields:
 *             self._pack(field)
 */
  __Pyx_INCREF(__pyx_v_write);
  __pyx_t_2 = __pyx_v_write; __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_6) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_signature); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_signature};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_signature};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6); __pyx_t_6 = NULL;
      __Pyx_INCREF(__pyx_v_signature);
      __Pyx_GIVEREF(__pyx_v_signature);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_signature);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":339
 *             raise OverflowError("Structure size out of range")
 *         write(signature)
 *         for field in fields:             # <<<<<<<<<<<<<<
 *             self._pack(field)
 * 
 */
  if (unlikely(__pyx_v_fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 339, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
  for (;;) {
    if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 339, __pyx_L1_error)
    #else
    __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "neo4j/packstream/_packer.pyx":340
 *         write(signature)
 *         for field in fields:
 *             self._pack(field)             # <<<<<<<<<<<<<<
 * 
 *     cpdef pack_end_of_stream(self):
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self->__pyx_vtab)->_pack(__pyx_v_self, __pyx_v_field); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "neo4j/packstream/_packer.pyx":339
 *             raise OverflowError("Structure size out of range")
 *         write(signature)
 *         for field in fields:             # <<<<<<<<<<<<<<
 *             self._pack(field)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":293
 *         self._write(b"\xDB")
 * 
 *     cpdef pack_struct(self, bytes signature, tuple fields):             # <<<<<<<<<<<<<<
 *         if len(signature) != 1:
 *             raise ValueError("Structure signature must be a single byte value")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_struct", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_write);
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_9pack_struct(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_9pack_struct(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_signature = 0;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_struct (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signature,&__pyx_n_s_fields,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_signature)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_fields)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pack_struct", 1, 2, 2, 1); __PYX_ERR(0, 293, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pack_struct") < 0)) __PYX_ERR(0, 293, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_signature = ((PyObject*)values[0]);
    __pyx_v_fields = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pack_struct", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 293, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_struct", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_signature), (&PyBytes_Type), 1, "signature", 1))) __PYX_ERR(0, 293, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_fields), (&PyTuple_Type), 1, "fields", 1))) __PYX_ERR(0, 293, __pyx_L1_error)
  __pyx_r = __pyx_pf_5neo4j_10packstream_7_packer_6Packer_8pack_struct(((struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self), __pyx_v_signature, __pyx_v_fields);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5neo4j_10packstream_7_packer_6Packer_8pack_struct(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_signature, PyObject *__pyx_v_fields) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pack_struct", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_struct(__pyx_v_self, __pyx_v_signature, __pyx_v_fields, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_struct", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "neo4j/packstream/_packer.pyx":342
 *             self._pack(field)
 * 
 *     cpdef pack_end_of_stream(self):             # <<<<<<<<<<<<<<
 *         self._write(b"\xDF")
 */

static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_11pack_end_of_stream(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_end_of_stream(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("pack_end_of_stream", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pack_end_of_stream); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5neo4j_10packstream_7_packer_6Packer_11pack_end_of_stream)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 342, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 342, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "neo4j/packstream/_packer.pyx":343
 * 
 *     cpdef pack_end_of_stream(self):
 *         self._write(b"\xDF")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_self->_write, __pyx_tuple__187, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":342
 *             self._pack(field)
 * 
 *     cpdef pack_end_of_stream(self):             # <<<<<<<<<<<<<<
 *         self._write(b"\xDF")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_end_of_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_11pack_end_of_stream(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_11pack_end_of_stream(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_end_of_stream (wrapper)", 0);
  __pyx_r = __pyx_pf_5neo4j_10packstream_7_packer_6Packer_10pack_end_of_stream(((struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5neo4j_10packstream_7_packer_6Packer_10pack_end_of_stream(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pack_end_of_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_end_of_stream(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.pack_end_of_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "neo4j/packstream/_packer.pyx":36
 * cdef class Packer(object):
 * 
 *     cdef public bint supports_bytes             # <<<<<<<<<<<<<<
 * 
 *     cdef stream
 */

/* Python wrapper */
static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_14supports_bytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5neo4j_10packstream_7_packer_6Packer_14supports_bytes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5neo4j_10packstream_7_packer_6Packer_14supports_bytes___get__(((struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5neo4j_10packstream_7_packer_6Packer_14supports_bytes___get__(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->supports_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.supports_bytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5neo4j_10packstream_7_packer_6Packer_14supports_bytes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5neo4j_10packstream_7_packer_6Packer_14supports_bytes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5neo4j_10packstream_7_packer_6Packer_14supports_bytes_2__set__(((struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5neo4j_10packstream_7_packer_6Packer_14supports_bytes_2__set__(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 36, __pyx_L1_error)
  __pyx_v_self->supports_bytes = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("neo4j.packstream._packer.Packer.supports_bytes.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_5neo4j_10packstream_7_packer_Packer __pyx_vtable_5neo4j_10packstream_7_packer_Packer;

static PyObject *__pyx_tp_new_5neo4j_10packstream_7_packer_Packer(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *)o);
  p->__pyx_vtab = __pyx_vtabptr_5neo4j_10packstream_7_packer_Packer;
  p->stream = Py_None; Py_INCREF(Py_None);
  p->_write = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_5neo4j_10packstream_7_packer_6Packer_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_5neo4j_10packstream_7_packer_Packer(PyObject *o) {
  struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *p = (struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->stream);
  Py_CLEAR(p->_write);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_5neo4j_10packstream_7_packer_Packer(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *p = (struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *)o;
  if (p->stream) {
    e = (*v)(p->stream, a); if (e) return e;
  }
  if (p->_write) {
    e = (*v)(p->_write, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5neo4j_10packstream_7_packer_Packer(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *p = (struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *)o;
  tmp = ((PyObject*)p->stream);
  p->stream = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_write);
  p->_write = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5neo4j_10packstream_7_packer_6Packer_supports_bytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5neo4j_10packstream_7_packer_6Packer_14supports_bytes_1__get__(o);
}

static int __pyx_setprop_5neo4j_10packstream_7_packer_6Packer_supports_bytes(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5neo4j_10packstream_7_packer_6Packer_14supports_bytes_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_5neo4j_10packstream_7_packer_Packer[] = {
  {"pack", (PyCFunction)__pyx_pw_5neo4j_10packstream_7_packer_6Packer_3pack, METH_O, 0},
  {"pack_list_stream_header", (PyCFunction)__pyx_pw_5neo4j_10packstream_7_packer_6Packer_5pack_list_stream_header, METH_NOARGS, 0},
  {"pack_map_stream_header", (PyCFunction)__pyx_pw_5neo4j_10packstream_7_packer_6Packer_7pack_map_stream_header, METH_NOARGS, 0},
  {"pack_struct", (PyCFunction)__pyx_pw_5neo4j_10packstream_7_packer_6Packer_9pack_struct, METH_VARARGS|METH_KEYWORDS, 0},
  {"pack_end_of_stream", (PyCFunction)__pyx_pw_5neo4j_10packstream_7_packer_6Packer_11pack_end_of_stream, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5neo4j_10packstream_7_packer_Packer[] = {
  {(char *)"supports_bytes", __pyx_getprop_5neo4j_10packstream_7_packer_6Packer_supports_bytes, __pyx_setprop_5neo4j_10packstream_7_packer_6Packer_supports_bytes, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5neo4j_10packstream_7_packer_Packer = {
  PyVarObject_HEAD_INIT(0, 0)
  "neo4j.packstream._packer.Packer", /*tp_name*/
  sizeof(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5neo4j_10packstream_7_packer_Packer, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5neo4j_10packstream_7_packer_Packer, /*tp_traverse*/
  __pyx_tp_clear_5neo4j_10packstream_7_packer_Packer, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5neo4j_10packstream_7_packer_Packer, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5neo4j_10packstream_7_packer_Packer, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5neo4j_10packstream_7_packer_Packer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

static int __pyx_import_star_set(PyObject *o, PyObject* py_name, char *name) {
  static const char* internal_type_names[] = {
    "Packer",
    0
  };
  const char** type_name = internal_type_names;
  while (*type_name) {
    if (__Pyx_StrEq(name, *type_name)) {
      PyErr_Format(PyExc_TypeError, "Cannot overwrite C type %s", name);
      goto bad;
    }
    type_name++;
  }
  if (0);
  else if (__Pyx_StrEq(name, "INT64_HI")) {
    Py_INCREF(o);
    Py_DECREF(__pyx_v_5neo4j_10packstream_7_packer_INT64_HI);
    __pyx_v_5neo4j_10packstream_7_packer_INT64_HI = o;
  }
  else if (__Pyx_StrEq(name, "INT64_LO")) {
    Py_INCREF(o);
    Py_DECREF(__pyx_v_5neo4j_10packstream_7_packer_INT64_LO);
    __pyx_v_5neo4j_10packstream_7_packer_INT64_LO = o;
  }
  else {
    if (PyObject_SetAttr(__pyx_m, py_name, o) < 0) goto bad;
  }
  return 0;
  bad:
  return -1;
}

























































































































static int
__Pyx_import_all_from(PyObject *locals, PyObject *v)
{
    PyObject *all = PyObject_GetAttrString(v, "__all__");
    PyObject *dict, *name, *value;
    int skip_leading_underscores = 0;
    int pos, err;

    if (all == NULL) {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError))
            return -1;
        PyErr_Clear();
        dict = PyObject_GetAttrString(v, "__dict__");
        if (dict == NULL) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError))
                return -1;
            PyErr_SetString(PyExc_ImportError,
            "from-import-* object has no __dict__ and no __all__");
            return -1;
        }
#if PY_MAJOR_VERSION < 3
        all = PyObject_CallMethod(dict, (char *)"keys", NULL);
#else
        all = PyMapping_Keys(dict);
#endif
        Py_DECREF(dict);
        if (all == NULL)
            return -1;
        skip_leading_underscores = 1;
    }

    for (pos = 0, err = 0; ; pos++) {
        name = PySequence_GetItem(all, pos);
        if (name == NULL) {
            if (!PyErr_ExceptionMatches(PyExc_IndexError))
                err = -1;
            else
                PyErr_Clear();
            break;
        }
        if (skip_leading_underscores &&
#if PY_MAJOR_VERSION < 3
            PyString_Check(name) &&
            PyString_AS_STRING(name)[0] == '_')
#else
            PyUnicode_Check(name) &&
            PyUnicode_AS_UNICODE(name)[0] == '_')
#endif
        {
            Py_DECREF(name);
            continue;
        }
        value = PyObject_GetAttr(v, name);
        if (value == NULL)
            err = -1;
        else if (PyDict_CheckExact(locals))
            err = PyDict_SetItem(locals, name, value);
        else
            err = PyObject_SetItem(locals, name, value);
        Py_DECREF(name);
        Py_XDECREF(value);
        if (err != 0)
            break;
    }
    Py_DECREF(all);
    return err;
}


static int __pyx_import_star(PyObject* m) {

    int i;
    int ret = -1;
    char* s;
    PyObject *locals = 0;
    PyObject *list = 0;
#if PY_MAJOR_VERSION >= 3
    PyObject *utf8_name = 0;
#endif
    PyObject *name;
    PyObject *item;

    locals = PyDict_New();              if (!locals) goto bad;
    if (__Pyx_import_all_from(locals, m) < 0) goto bad;
    list = PyDict_Items(locals);        if (!list) goto bad;

    for(i=0; i<PyList_GET_SIZE(list); i++) {
        name = PyTuple_GET_ITEM(PyList_GET_ITEM(list, i), 0);
        item = PyTuple_GET_ITEM(PyList_GET_ITEM(list, i), 1);
#if PY_MAJOR_VERSION >= 3
        utf8_name = PyUnicode_AsUTF8String(name);
        if (!utf8_name) goto bad;
        s = PyBytes_AS_STRING(utf8_name);
        if (__pyx_import_star_set(item, name, s) < 0) goto bad;
        Py_DECREF(utf8_name); utf8_name = 0;
#else
        s = PyString_AsString(name);
        if (!s) goto bad;
        if (__pyx_import_star_set(item, name, s) < 0) goto bad;
#endif
    }
    ret = 0;

bad:
    Py_XDECREF(locals);
    Py_XDECREF(list);
#if PY_MAJOR_VERSION >= 3
    Py_XDECREF(utf8_name);
#endif
    return ret;
}



#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "_packer",
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_b_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 0, 0},
  {&__pyx_kp_s_Bytes_header_size_out_of_range, __pyx_k_Bytes_header_size_out_of_range, sizeof(__pyx_k_Bytes_header_size_out_of_range), 0, 0, 1, 0},
  {&__pyx_kp_s_I, __pyx_k_I, sizeof(__pyx_k_I), 0, 0, 1, 0},
  {&__pyx_kp_s_Integer_s_out_of_range, __pyx_k_Integer_s_out_of_range, sizeof(__pyx_k_Integer_s_out_of_range), 0, 0, 1, 0},
  {&__pyx_kp_s_List_header_size_out_of_range, __pyx_k_List_header_size_out_of_range, sizeof(__pyx_k_List_header_size_out_of_range), 0, 0, 1, 0},
  {&__pyx_kp_s_Map_header_size_out_of_range, __pyx_k_Map_header_size_out_of_range, sizeof(__pyx_k_Map_header_size_out_of_range), 0, 0, 1, 0},
  {&__pyx_n_s_OverflowError, __pyx_k_OverflowError, sizeof(__pyx_k_OverflowError), 0, 0, 1, 1},
  {&__pyx_n_s_PACKED_UINT_16, __pyx_k_PACKED_UINT_16, sizeof(__pyx_k_PACKED_UINT_16), 0, 0, 1, 1},
  {&__pyx_n_s_PACKED_UINT_8, __pyx_k_PACKED_UINT_8, sizeof(__pyx_k_PACKED_UINT_8), 0, 0, 1, 1},
  {&__pyx_kp_s_String_header_size_out_of_range, __pyx_k_String_header_size_out_of_range, sizeof(__pyx_k_String_header_size_out_of_range), 0, 0, 1, 0},
  {&__pyx_n_s_Structure, __pyx_k_Structure, sizeof(__pyx_k_Structure), 0, 0, 1, 1},
  {&__pyx_kp_s_Structure_signature_must_be_a_si, __pyx_k_Structure_signature_must_be_a_si, sizeof(__pyx_k_Structure_signature_must_be_a_si), 0, 0, 1, 0},
  {&__pyx_kp_s_Structure_size_out_of_range, __pyx_k_Structure_size_out_of_range, sizeof(__pyx_k_Structure_size_out_of_range), 0, 0, 1, 0},
  {&__pyx_kp_s_Structures_require_a_2_tuple_of, __pyx_k_Structures_require_a_2_tuple_of, sizeof(__pyx_k_Structures_require_a_2_tuple_of), 0, 0, 1, 0},
  {&__pyx_kp_s_This_PackSteam_channel_does_not, __pyx_k_This_PackSteam_channel_does_not, sizeof(__pyx_k_This_PackSteam_channel_does_not), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s_Values_of_type_s_are_not_support, __pyx_k_Values_of_type_s_are_not_support, sizeof(__pyx_k_Values_of_type_s_are_not_support), 0, 0, 1, 0},
  {&__pyx_kp_b__100, __pyx_k__100, sizeof(__pyx_k__100), 0, 0, 0, 0},
  {&__pyx_kp_b__102, __pyx_k__102, sizeof(__pyx_k__102), 0, 0, 0, 0},
  {&__pyx_kp_b__105, __pyx_k__105, sizeof(__pyx_k__105), 0, 0, 0, 0},
  {&__pyx_kp_b__107, __pyx_k__107, sizeof(__pyx_k__107), 0, 0, 0, 0},
  {&__pyx_kp_b__109, __pyx_k__109, sizeof(__pyx_k__109), 0, 0, 0, 0},
  {&__pyx_kp_b__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 0, 0, 0},
  {&__pyx_kp_b__111, __pyx_k__111, sizeof(__pyx_k__111), 0, 0, 0, 0},
  {&__pyx_kp_b__113, __pyx_k__113, sizeof(__pyx_k__113), 0, 0, 0, 0},
  {&__pyx_kp_b__115, __pyx_k__115, sizeof(__pyx_k__115), 0, 0, 0, 0},
  {&__pyx_kp_b__117, __pyx_k__117, sizeof(__pyx_k__117), 0, 0, 0, 0},
  {&__pyx_kp_b__119, __pyx_k__119, sizeof(__pyx_k__119), 0, 0, 0, 0},
  {&__pyx_kp_b__121, __pyx_k__121, sizeof(__pyx_k__121), 0, 0, 0, 0},
  {&__pyx_kp_b__123, __pyx_k__123, sizeof(__pyx_k__123), 0, 0, 0, 0},
  {&__pyx_kp_b__125, __pyx_k__125, sizeof(__pyx_k__125), 0, 0, 0, 0},
  {&__pyx_kp_b__127, __pyx_k__127, sizeof(__pyx_k__127), 0, 0, 0, 0},
  {&__pyx_kp_b__129, __pyx_k__129, sizeof(__pyx_k__129), 0, 0, 0, 0},
  {&__pyx_kp_b__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 0, 0, 0},
  {&__pyx_kp_b__131, __pyx_k__131, sizeof(__pyx_k__131), 0, 0, 0, 0},
  {&__pyx_kp_b__133, __pyx_k__133, sizeof(__pyx_k__133), 0, 0, 0, 0},
  {&__pyx_kp_b__135, __pyx_k__135, sizeof(__pyx_k__135), 0, 0, 0, 0},
  {&__pyx_kp_b__137, __pyx_k__137, sizeof(__pyx_k__137), 0, 0, 0, 0},
  {&__pyx_kp_b__139, __pyx_k__139, sizeof(__pyx_k__139), 0, 0, 0, 0},
  {&__pyx_kp_b__141, __pyx_k__141, sizeof(__pyx_k__141), 0, 0, 0, 0},
  {&__pyx_kp_b__143, __pyx_k__143, sizeof(__pyx_k__143), 0, 0, 0, 0},
  {&__pyx_kp_b__146, __pyx_k__146, sizeof(__pyx_k__146), 0, 0, 0, 0},
  {&__pyx_kp_b__149, __pyx_k__149, sizeof(__pyx_k__149), 0, 0, 0, 0},
  {&__pyx_kp_b__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 0, 0, 0},
  {&__pyx_kp_b__151, __pyx_k__151, sizeof(__pyx_k__151), 0, 0, 0, 0},
  {&__pyx_kp_b__153, __pyx_k__153, sizeof(__pyx_k__153), 0, 0, 0, 0},
  {&__pyx_kp_b__155, __pyx_k__155, sizeof(__pyx_k__155), 0, 0, 0, 0},
  {&__pyx_kp_b__157, __pyx_k__157, sizeof(__pyx_k__157), 0, 0, 0, 0},
  {&__pyx_kp_b__159, __pyx_k__159, sizeof(__pyx_k__159), 0, 0, 0, 0},
  {&__pyx_kp_b__161, __pyx_k__161, sizeof(__pyx_k__161), 0, 0, 0, 0},
  {&__pyx_kp_b__163, __pyx_k__163, sizeof(__pyx_k__163), 0, 0, 0, 0},
  {&__pyx_kp_b__165, __pyx_k__165, sizeof(__pyx_k__165), 0, 0, 0, 0},
  {&__pyx_kp_b__167, __pyx_k__167, sizeof(__pyx_k__167), 0, 0, 0, 0},
  {&__pyx_kp_b__169, __pyx_k__169, sizeof(__pyx_k__169), 0, 0, 0, 0},
  {&__pyx_kp_b__171, __pyx_k__171, sizeof(__pyx_k__171), 0, 0, 0, 0},
  {&__pyx_kp_b__173, __pyx_k__173, sizeof(__pyx_k__173), 0, 0, 0, 0},
  {&__pyx_kp_b__175, __pyx_k__175, sizeof(__pyx_k__175), 0, 0, 0, 0},
  {&__pyx_kp_b__177, __pyx_k__177, sizeof(__pyx_k__177), 0, 0, 0, 0},
  {&__pyx_kp_b__179, __pyx_k__179, sizeof(__pyx_k__179), 0, 0, 0, 0},
  {&__pyx_kp_b__181, __pyx_k__181, sizeof(__pyx_k__181), 0, 0, 0, 0},
  {&__pyx_kp_b__183, __pyx_k__183, sizeof(__pyx_k__183), 0, 0, 0, 0},
  {&__pyx_kp_b__186, __pyx_k__186, sizeof(__pyx_k__186), 0, 0, 0, 0},
  {&__pyx_n_s__188, __pyx_k__188, sizeof(__pyx_k__188), 0, 0, 1, 1},
  {&__pyx_kp_b__20, __pyx_k__20, sizeof(__pyx_k__20), 0, 0, 0, 0},
  {&__pyx_kp_b__22, __pyx_k__22, sizeof(__pyx_k__22), 0, 0, 0, 0},
  {&__pyx_kp_b__24, __pyx_k__24, sizeof(__pyx_k__24), 0, 0, 0, 0},
  {&__pyx_kp_b__27, __pyx_k__27, sizeof(__pyx_k__27), 0, 0, 0, 0},
  {&__pyx_kp_b__29, __pyx_k__29, sizeof(__pyx_k__29), 0, 0, 0, 0},
  {&__pyx_kp_b__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 0, 0},
  {&__pyx_kp_b__31, __pyx_k__31, sizeof(__pyx_k__31), 0, 0, 0, 0},
  {&__pyx_kp_b__33, __pyx_k__33, sizeof(__pyx_k__33), 0, 0, 0, 0},
  {&__pyx_kp_b__35, __pyx_k__35, sizeof(__pyx_k__35), 0, 0, 0, 0},
  {&__pyx_kp_b__37, __pyx_k__37, sizeof(__pyx_k__37), 0, 0, 0, 0},
  {&__pyx_kp_b__39, __pyx_k__39, sizeof(__pyx_k__39), 0, 0, 0, 0},
  {&__pyx_kp_b__41, __pyx_k__41, sizeof(__pyx_k__41), 0, 0, 0, 0},
  {&__pyx_kp_b__43, __pyx_k__43, sizeof(__pyx_k__43), 0, 0, 0, 0},
  {&__pyx_kp_b__45, __pyx_k__45, sizeof(__pyx_k__45), 0, 0, 0, 0},
  {&__pyx_kp_b__47, __pyx_k__47, sizeof(__pyx_k__47), 0, 0, 0, 0},
  {&__pyx_kp_b__49, __pyx_k__49, sizeof(__pyx_k__49), 0, 0, 0, 0},
  {&__pyx_kp_b__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 0, 0},
  {&__pyx_kp_b__51, __pyx_k__51, sizeof(__pyx_k__51), 0, 0, 0, 0},
  {&__pyx_kp_b__53, __pyx_k__53, sizeof(__pyx_k__53), 0, 0, 0, 0},
  {&__pyx_kp_b__55, __pyx_k__55, sizeof(__pyx_k__55), 0, 0, 0, 0},
  {&__pyx_kp_b__57, __pyx_k__57, sizeof(__pyx_k__57), 0, 0, 0, 0},
  {&__pyx_kp_b__59, __pyx_k__59, sizeof(__pyx_k__59), 0, 0, 0, 0},
  {&__pyx_kp_b__61, __pyx_k__61, sizeof(__pyx_k__61), 0, 0, 0, 0},
  {&__pyx_kp_b__63, __pyx_k__63, sizeof(__pyx_k__63), 0, 0, 0, 0},
  {&__pyx_kp_b__66, __pyx_k__66, sizeof(__pyx_k__66), 0, 0, 0, 0},
  {&__pyx_kp_b__68, __pyx_k__68, sizeof(__pyx_k__68), 0, 0, 0, 0},
  {&__pyx_kp_b__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 0, 0, 0},
  {&__pyx_kp_b__70, __pyx_k__70, sizeof(__pyx_k__70), 0, 0, 0, 0},
  {&__pyx_kp_b__72, __pyx_k__72, sizeof(__pyx_k__72), 0, 0, 0, 0},
  {&__pyx_kp_b__74, __pyx_k__74, sizeof(__pyx_k__74), 0, 0, 0, 0},
  {&__pyx_kp_b__76, __pyx_k__76, sizeof(__pyx_k__76), 0, 0, 0, 0},
  {&__pyx_kp_b__78, __pyx_k__78, sizeof(__pyx_k__78), 0, 0, 0, 0},
  {&__pyx_kp_b__80, __pyx_k__80, sizeof(__pyx_k__80), 0, 0, 0, 0},
  {&__pyx_kp_b__82, __pyx_k__82, sizeof(__pyx_k__82), 0, 0, 0, 0},
  {&__pyx_kp_b__84, __pyx_k__84, sizeof(__pyx_k__84), 0, 0, 0, 0},
  {&__pyx_kp_b__86, __pyx_k__86, sizeof(__pyx_k__86), 0, 0, 0, 0},
  {&__pyx_kp_b__88, __pyx_k__88, sizeof(__pyx_k__88), 0, 0, 0, 0},
  {&__pyx_kp_b__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 0, 0, 0},
  {&__pyx_kp_b__90, __pyx_k__90, sizeof(__pyx_k__90), 0, 0, 0, 0},
  {&__pyx_kp_b__92, __pyx_k__92, sizeof(__pyx_k__92), 0, 0, 0, 0},
  {&__pyx_kp_b__94, __pyx_k__94, sizeof(__pyx_k__94), 0, 0, 0, 0},
  {&__pyx_kp_b__96, __pyx_k__96, sizeof(__pyx_k__96), 0, 0, 0, 0},
  {&__pyx_kp_b__98, __pyx_k__98, sizeof(__pyx_k__98), 0, 0, 0, 0},
  {&__pyx_n_s_constants, __pyx_k_constants, sizeof(__pyx_k_constants), 0, 0, 1, 1},
  {&__pyx_kp_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 0},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_fields, __pyx_k_fields, sizeof(__pyx_k_fields), 0, 0, 1, 1},
  {&__pyx_kp_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_integer, __pyx_k_integer, sizeof(__pyx_k_integer), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_neo4j_compat, __pyx_k_neo4j_compat, sizeof(__pyx_k_neo4j_compat), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_pack_end_of_stream, __pyx_k_pack_end_of_stream, sizeof(__pyx_k_pack_end_of_stream), 0, 0, 1, 1},
  {&__pyx_n_s_pack_list_stream_header, __pyx_k_pack_list_stream_header, sizeof(__pyx_k_pack_list_stream_header), 0, 0, 1, 1},
  {&__pyx_n_s_pack_map_stream_header, __pyx_k_pack_map_stream_header, sizeof(__pyx_k_pack_map_stream_header), 0, 0, 1, 1},
  {&__pyx_n_s_pack_struct, __pyx_k_pack_struct, sizeof(__pyx_k_pack_struct), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_kp_s_q, __pyx_k_q, sizeof(__pyx_k_q), 0, 0, 1, 0},
  {&__pyx_n_s_signature, __pyx_k_signature, sizeof(__pyx_k_signature), 0, 0, 1, 1},
  {&__pyx_n_s_stream, __pyx_k_stream, sizeof(__pyx_k_stream), 0, 0, 1, 1},
  {&__pyx_n_s_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 0, 1, 1},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_struct_pack, __pyx_k_struct_pack, sizeof(__pyx_k_struct_pack), 0, 0, 1, 1},
  {&__pyx_n_s_structure, __pyx_k_structure, sizeof(__pyx_k_structure), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_s_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 0, 1, 0},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_OverflowError = __Pyx_GetBuiltinName(__pyx_n_s_OverflowError); if (!__pyx_builtin_OverflowError) __PYX_ERR(0, 87, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 124, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 135, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "neo4j/packstream/_packer.pyx":57
 *         # None
 *         if value is None:
 *             write(b"\xC0")  # NULL             # <<<<<<<<<<<<<<
 * 
 *         # Boolean
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_b_); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "neo4j/packstream/_packer.pyx":61
 *         # Boolean
 *         elif value is True:
 *             write(b"\xC3")             # <<<<<<<<<<<<<<
 *         elif value is False:
 *             write(b"\xC2")
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_b__3); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "neo4j/packstream/_packer.pyx":63
 *             write(b"\xC3")
 *         elif value is False:
 *             write(b"\xC2")             # <<<<<<<<<<<<<<
 * 
 *         # Float (only double precision is supported)
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_b__5); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "neo4j/packstream/_packer.pyx":67
 *         # Float (only double precision is supported)
 *         elif isinstance(value, float):
 *             write(b"\xC1")             # <<<<<<<<<<<<<<
 *             write(struct_pack(">d", value))
 * 
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_b__7); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "neo4j/packstream/_packer.pyx":75
 *                 write(PACKED_UINT_8[value % 0x100])
 *             elif -0x80 <= value < -0x10:
 *                 write(b"\xC8")             # <<<<<<<<<<<<<<
 *                 write(PACKED_UINT_8[value % 0x100])
 *             elif -0x8000 <= value < 0x8000:
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_b__9); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "neo4j/packstream/_packer.pyx":78
 *                 write(PACKED_UINT_8[value % 0x100])
 *             elif -0x8000 <= value < 0x8000:
 *                 write(b"\xC9")             # <<<<<<<<<<<<<<
 *                 write(PACKED_UINT_16[value % 0x10000])
 *             elif -0x80000000 <= value < 0x80000000:
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_b__11); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "neo4j/packstream/_packer.pyx":81
 *                 write(PACKED_UINT_16[value % 0x10000])
 *             elif -0x80000000 <= value < 0x80000000:
 *                 write(b"\xCA")             # <<<<<<<<<<<<<<
 *                 write(struct_pack(">i", value))
 *             elif INT64_LO <= value < INT64_HI:
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_b__13); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "neo4j/packstream/_packer.pyx":84
 *                 write(struct_pack(">i", value))
 *             elif INT64_LO <= value < INT64_HI:
 *                 write(b"\xCB")             # <<<<<<<<<<<<<<
 *                 write(struct_pack(">q", value))
 *             else:
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_b__15); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "neo4j/packstream/_packer.pyx":94
 *                 value_bytes = value
 *             else:
 *                 value_bytes = value.encode("utf-8")             # <<<<<<<<<<<<<<
 *             self.pack_string_header(len(value_bytes))
 *             self.pack_raw(value_bytes)
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "neo4j/packstream/_packer.pyx":125
 *                 signature, fields = value
 *             except ValueError:
 *                 raise ValueError("Structures require a 2-tuple of (signature, fields)")             # <<<<<<<<<<<<<<
 *             else:
 *                 self.pack_struct(signature, fields)
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_Structures_require_a_2_tuple_of); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "neo4j/packstream/_packer.pyx":135
 *     cdef pack_bytes_header(self, int size):
 *         if not self.supports_bytes:
 *             raise TypeError("This PackSteam channel does not support BYTES (consider upgrading to Neo4j 3.2+)")             # <<<<<<<<<<<<<<
 *         write = self._write
 *         if size < 0x100:
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_This_PackSteam_channel_does_not); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "neo4j/packstream/_packer.pyx":138
 *         write = self._write
 *         if size < 0x100:
 *             write(b"\xCC")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_b__20); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "neo4j/packstream/_packer.pyx":141
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 *             write(b"\xCD")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_b__22); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "neo4j/packstream/_packer.pyx":144
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 *             write(b"\xCE")             # <<<<<<<<<<<<<<
 *             write(struct_pack(">I", size))
 *         else:
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_b__24); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "neo4j/packstream/_packer.pyx":147
 *             write(struct_pack(">I", size))
 *         else:
 *             raise OverflowError("Bytes header size out of range")             # <<<<<<<<<<<<<<
 * 
 *     cdef pack_string_header(self, int size):
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_s_Bytes_header_size_out_of_range); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "neo4j/packstream/_packer.pyx":152
 *         write = self._write
 *         if size == 0x00:
 *             write(b"\x80")             # <<<<<<<<<<<<<<
 *         elif size == 0x01:
 *             write(b"\x81")
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_b__27); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "neo4j/packstream/_packer.pyx":154
 *             write(b"\x80")
 *         elif size == 0x01:
 *             write(b"\x81")             # <<<<<<<<<<<<<<
 *         elif size == 0x02:
 *             write(b"\x82")
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_b__29); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "neo4j/packstream/_packer.pyx":156
 *             write(b"\x81")
 *         elif size == 0x02:
 *             write(b"\x82")             # <<<<<<<<<<<<<<
 *         elif size == 0x03:
 *             write(b"\x83")
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_b__31); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "neo4j/packstream/_packer.pyx":158
 *             write(b"\x82")
 *         elif size == 0x03:
 *             write(b"\x83")             # <<<<<<<<<<<<<<
 *         elif size == 0x04:
 *             write(b"\x84")
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_b__33); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "neo4j/packstream/_packer.pyx":160
 *             write(b"\x83")
 *         elif size == 0x04:
 *             write(b"\x84")             # <<<<<<<<<<<<<<
 *         elif size == 0x05:
 *             write(b"\x85")
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_b__35); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "neo4j/packstream/_packer.pyx":162
 *             write(b"\x84")
 *         elif size == 0x05:
 *             write(b"\x85")             # <<<<<<<<<<<<<<
 *         elif size == 0x06:
 *             write(b"\x86")
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_kp_b__37); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "neo4j/packstream/_packer.pyx":164
 *             write(b"\x85")
 *         elif size == 0x06:
 *             write(b"\x86")             # <<<<<<<<<<<<<<
 *         elif size == 0x07:
 *             write(b"\x87")
 */
  __pyx_tuple__40 = PyTuple_Pack(1, __pyx_kp_b__39); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "neo4j/packstream/_packer.pyx":166
 *             write(b"\x86")
 *         elif size == 0x07:
 *             write(b"\x87")             # <<<<<<<<<<<<<<
 *         elif size == 0x08:
 *             write(b"\x88")
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_kp_b__41); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "neo4j/packstream/_packer.pyx":168
 *             write(b"\x87")
 *         elif size == 0x08:
 *             write(b"\x88")             # <<<<<<<<<<<<<<
 *         elif size == 0x09:
 *             write(b"\x89")
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_kp_b__43); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "neo4j/packstream/_packer.pyx":170
 *             write(b"\x88")
 *         elif size == 0x09:
 *             write(b"\x89")             # <<<<<<<<<<<<<<
 *         elif size == 0x0A:
 *             write(b"\x8A")
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_kp_b__45); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "neo4j/packstream/_packer.pyx":172
 *             write(b"\x89")
 *         elif size == 0x0A:
 *             write(b"\x8A")             # <<<<<<<<<<<<<<
 *         elif size == 0x0B:
 *             write(b"\x8B")
 */
  __pyx_tuple__48 = PyTuple_Pack(1, __pyx_kp_b__47); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "neo4j/packstream/_packer.pyx":174
 *             write(b"\x8A")
 *         elif size == 0x0B:
 *             write(b"\x8B")             # <<<<<<<<<<<<<<
 *         elif size == 0x0C:
 *             write(b"\x8C")
 */
  __pyx_tuple__50 = PyTuple_Pack(1, __pyx_kp_b__49); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "neo4j/packstream/_packer.pyx":176
 *             write(b"\x8B")
 *         elif size == 0x0C:
 *             write(b"\x8C")             # <<<<<<<<<<<<<<
 *         elif size == 0x0D:
 *             write(b"\x8D")
 */
  __pyx_tuple__52 = PyTuple_Pack(1, __pyx_kp_b__51); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);

  /* "neo4j/packstream/_packer.pyx":178
 *             write(b"\x8C")
 *         elif size == 0x0D:
 *             write(b"\x8D")             # <<<<<<<<<<<<<<
 *         elif size == 0x0E:
 *             write(b"\x8E")
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_kp_b__53); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "neo4j/packstream/_packer.pyx":180
 *             write(b"\x8D")
 *         elif size == 0x0E:
 *             write(b"\x8E")             # <<<<<<<<<<<<<<
 *         elif size == 0x0F:
 *             write(b"\x8F")
 */
  __pyx_tuple__56 = PyTuple_Pack(1, __pyx_kp_b__55); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "neo4j/packstream/_packer.pyx":182
 *             write(b"\x8E")
 *         elif size == 0x0F:
 *             write(b"\x8F")             # <<<<<<<<<<<<<<
 *         elif size < 0x100:
 *             write(b"\xD0")
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_kp_b__57); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "neo4j/packstream/_packer.pyx":184
 *             write(b"\x8F")
 *         elif size < 0x100:
 *             write(b"\xD0")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 */
  __pyx_tuple__60 = PyTuple_Pack(1, __pyx_kp_b__59); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "neo4j/packstream/_packer.pyx":187
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 *             write(b"\xD1")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 */
  __pyx_tuple__62 = PyTuple_Pack(1, __pyx_kp_b__61); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "neo4j/packstream/_packer.pyx":190
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 *             write(b"\xD2")             # <<<<<<<<<<<<<<
 *             write(struct_pack(">I", size))
 *         else:
 */
  __pyx_tuple__64 = PyTuple_Pack(1, __pyx_kp_b__63); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);

  /* "neo4j/packstream/_packer.pyx":193
 *             write(struct_pack(">I", size))
 *         else:
 *             raise OverflowError("String header size out of range")             # <<<<<<<<<<<<<<
 * 
 *     cdef pack_list_header(self, int size):
 */
  __pyx_tuple__65 = PyTuple_Pack(1, __pyx_kp_s_String_header_size_out_of_range); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "neo4j/packstream/_packer.pyx":198
 *         write = self._write
 *         if size == 0x00:
 *             write(b"\x90")             # <<<<<<<<<<<<<<
 *         elif size == 0x01:
 *             write(b"\x91")
 */
  __pyx_tuple__67 = PyTuple_Pack(1, __pyx_kp_b__66); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);

  /* "neo4j/packstream/_packer.pyx":200
 *             write(b"\x90")
 *         elif size == 0x01:
 *             write(b"\x91")             # <<<<<<<<<<<<<<
 *         elif size == 0x02:
 *             write(b"\x92")
 */
  __pyx_tuple__69 = PyTuple_Pack(1, __pyx_kp_b__68); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);

  /* "neo4j/packstream/_packer.pyx":202
 *             write(b"\x91")
 *         elif size == 0x02:
 *             write(b"\x92")             # <<<<<<<<<<<<<<
 *         elif size == 0x03:
 *             write(b"\x93")
 */
  __pyx_tuple__71 = PyTuple_Pack(1, __pyx_kp_b__70); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);

  /* "neo4j/packstream/_packer.pyx":204
 *             write(b"\x92")
 *         elif size == 0x03:
 *             write(b"\x93")             # <<<<<<<<<<<<<<
 *         elif size == 0x04:
 *             write(b"\x94")
 */
  __pyx_tuple__73 = PyTuple_Pack(1, __pyx_kp_b__72); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);

  /* "neo4j/packstream/_packer.pyx":206
 *             write(b"\x93")
 *         elif size == 0x04:
 *             write(b"\x94")             # <<<<<<<<<<<<<<
 *         elif size == 0x05:
 *             write(b"\x95")
 */
  __pyx_tuple__75 = PyTuple_Pack(1, __pyx_kp_b__74); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);

  /* "neo4j/packstream/_packer.pyx":208
 *             write(b"\x94")
 *         elif size == 0x05:
 *             write(b"\x95")             # <<<<<<<<<<<<<<
 *         elif size == 0x06:
 *             write(b"\x96")
 */
  __pyx_tuple__77 = PyTuple_Pack(1, __pyx_kp_b__76); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);

  /* "neo4j/packstream/_packer.pyx":210
 *             write(b"\x95")
 *         elif size == 0x06:
 *             write(b"\x96")             # <<<<<<<<<<<<<<
 *         elif size == 0x07:
 *             write(b"\x97")
 */
  __pyx_tuple__79 = PyTuple_Pack(1, __pyx_kp_b__78); if (unlikely(!__pyx_tuple__79)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);

  /* "neo4j/packstream/_packer.pyx":212
 *             write(b"\x96")
 *         elif size == 0x07:
 *             write(b"\x97")             # <<<<<<<<<<<<<<
 *         elif size == 0x08:
 *             write(b"\x98")
 */
  __pyx_tuple__81 = PyTuple_Pack(1, __pyx_kp_b__80); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);

  /* "neo4j/packstream/_packer.pyx":214
 *             write(b"\x97")
 *         elif size == 0x08:
 *             write(b"\x98")             # <<<<<<<<<<<<<<
 *         elif size == 0x09:
 *             write(b"\x99")
 */
  __pyx_tuple__83 = PyTuple_Pack(1, __pyx_kp_b__82); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);

  /* "neo4j/packstream/_packer.pyx":216
 *             write(b"\x98")
 *         elif size == 0x09:
 *             write(b"\x99")             # <<<<<<<<<<<<<<
 *         elif size == 0x0A:
 *             write(b"\x9A")
 */
  __pyx_tuple__85 = PyTuple_Pack(1, __pyx_kp_b__84); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);

  /* "neo4j/packstream/_packer.pyx":218
 *             write(b"\x99")
 *         elif size == 0x0A:
 *             write(b"\x9A")             # <<<<<<<<<<<<<<
 *         elif size == 0x0B:
 *             write(b"\x9B")
 */
  __pyx_tuple__87 = PyTuple_Pack(1, __pyx_kp_b__86); if (unlikely(!__pyx_tuple__87)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);

  /* "neo4j/packstream/_packer.pyx":220
 *             write(b"\x9A")
 *         elif size == 0x0B:
 *             write(b"\x9B")             # <<<<<<<<<<<<<<
 *         elif size == 0x0C:
 *             write(b"\x9C")
 */
  __pyx_tuple__89 = PyTuple_Pack(1, __pyx_kp_b__88); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);

  /* "neo4j/packstream/_packer.pyx":222
 *             write(b"\x9B")
 *         elif size == 0x0C:
 *             write(b"\x9C")             # <<<<<<<<<<<<<<
 *         elif size == 0x0D:
 *             write(b"\x9D")
 */
  __pyx_tuple__91 = PyTuple_Pack(1, __pyx_kp_b__90); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);

  /* "neo4j/packstream/_packer.pyx":224
 *             write(b"\x9C")
 *         elif size == 0x0D:
 *             write(b"\x9D")             # <<<<<<<<<<<<<<
 *         elif size == 0x0E:
 *             write(b"\x9E")
 */
  __pyx_tuple__93 = PyTuple_Pack(1, __pyx_kp_b__92); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);

  /* "neo4j/packstream/_packer.pyx":226
 *             write(b"\x9D")
 *         elif size == 0x0E:
 *             write(b"\x9E")             # <<<<<<<<<<<<<<
 *         elif size == 0x0F:
 *             write(b"\x9F")
 */
  __pyx_tuple__95 = PyTuple_Pack(1, __pyx_kp_b__94); if (unlikely(!__pyx_tuple__95)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);

  /* "neo4j/packstream/_packer.pyx":228
 *             write(b"\x9E")
 *         elif size == 0x0F:
 *             write(b"\x9F")             # <<<<<<<<<<<<<<
 *         elif size < 0x100:
 *             write(b"\xD4")
 */
  __pyx_tuple__97 = PyTuple_Pack(1, __pyx_kp_b__96); if (unlikely(!__pyx_tuple__97)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__97);
  __Pyx_GIVEREF(__pyx_tuple__97);

  /* "neo4j/packstream/_packer.pyx":230
 *             write(b"\x9F")
 *         elif size < 0x100:
 *             write(b"\xD4")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 */
  __pyx_tuple__99 = PyTuple_Pack(1, __pyx_kp_b__98); if (unlikely(!__pyx_tuple__99)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);

  /* "neo4j/packstream/_packer.pyx":233
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 *             write(b"\xD5")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 */
  __pyx_tuple__101 = PyTuple_Pack(1, __pyx_kp_b__100); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);

  /* "neo4j/packstream/_packer.pyx":236
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 *             write(b"\xD6")             # <<<<<<<<<<<<<<
 *             write(struct_pack(">I", size))
 *         else:
 */
  __pyx_tuple__103 = PyTuple_Pack(1, __pyx_kp_b__102); if (unlikely(!__pyx_tuple__103)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);

  /* "neo4j/packstream/_packer.pyx":239
 *             write(struct_pack(">I", size))
 *         else:
 *             raise OverflowError("List header size out of range")             # <<<<<<<<<<<<<<
 * 
 *     cpdef pack_list_stream_header(self):
 */
  __pyx_tuple__104 = PyTuple_Pack(1, __pyx_kp_s_List_header_size_out_of_range); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);

  /* "neo4j/packstream/_packer.pyx":242
 * 
 *     cpdef pack_list_stream_header(self):
 *         self._write(b"\xD7")             # <<<<<<<<<<<<<<
 * 
 *     cdef pack_map_header(self, int size):
 */
  __pyx_tuple__106 = PyTuple_Pack(1, __pyx_kp_b__105); if (unlikely(!__pyx_tuple__106)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);

  /* "neo4j/packstream/_packer.pyx":247
 *         write = self._write
 *         if size == 0x00:
 *             write(b"\xA0")             # <<<<<<<<<<<<<<
 *         elif size == 0x01:
 *             write(b"\xA1")
 */
  __pyx_tuple__108 = PyTuple_Pack(1, __pyx_kp_b__107); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);

  /* "neo4j/packstream/_packer.pyx":249
 *             write(b"\xA0")
 *         elif size == 0x01:
 *             write(b"\xA1")             # <<<<<<<<<<<<<<
 *         elif size == 0x02:
 *             write(b"\xA2")
 */
  __pyx_tuple__110 = PyTuple_Pack(1, __pyx_kp_b__109); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);

  /* "neo4j/packstream/_packer.pyx":251
 *             write(b"\xA1")
 *         elif size == 0x02:
 *             write(b"\xA2")             # <<<<<<<<<<<<<<
 *         elif size == 0x03:
 *             write(b"\xA3")
 */
  __pyx_tuple__112 = PyTuple_Pack(1, __pyx_kp_b__111); if (unlikely(!__pyx_tuple__112)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);

  /* "neo4j/packstream/_packer.pyx":253
 *             write(b"\xA2")
 *         elif size == 0x03:
 *             write(b"\xA3")             # <<<<<<<<<<<<<<
 *         elif size == 0x04:
 *             write(b"\xA4")
 */
  __pyx_tuple__114 = PyTuple_Pack(1, __pyx_kp_b__113); if (unlikely(!__pyx_tuple__114)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);

  /* "neo4j/packstream/_packer.pyx":255
 *             write(b"\xA3")
 *         elif size == 0x04:
 *             write(b"\xA4")             # <<<<<<<<<<<<<<
 *         elif size == 0x05:
 *             write(b"\xA5")
 */
  __pyx_tuple__116 = PyTuple_Pack(1, __pyx_kp_b__115); if (unlikely(!__pyx_tuple__116)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);

  /* "neo4j/packstream/_packer.pyx":257
 *             write(b"\xA4")
 *         elif size == 0x05:
 *             write(b"\xA5")             # <<<<<<<<<<<<<<
 *         elif size == 0x06:
 *             write(b"\xA6")
 */
  __pyx_tuple__118 = PyTuple_Pack(1, __pyx_kp_b__117); if (unlikely(!__pyx_tuple__118)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);

  /* "neo4j/packstream/_packer.pyx":259
 *             write(b"\xA5")
 *         elif size == 0x06:
 *             write(b"\xA6")             # <<<<<<<<<<<<<<
 *         elif size == 0x07:
 *             write(b"\xA7")
 */
  __pyx_tuple__120 = PyTuple_Pack(1, __pyx_kp_b__119); if (unlikely(!__pyx_tuple__120)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);

  /* "neo4j/packstream/_packer.pyx":261
 *             write(b"\xA6")
 *         elif size == 0x07:
 *             write(b"\xA7")             # <<<<<<<<<<<<<<
 *         elif size == 0x08:
 *             write(b"\xA8")
 */
  __pyx_tuple__122 = PyTuple_Pack(1, __pyx_kp_b__121); if (unlikely(!__pyx_tuple__122)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);

  /* "neo4j/packstream/_packer.pyx":263
 *             write(b"\xA7")
 *         elif size == 0x08:
 *             write(b"\xA8")             # <<<<<<<<<<<<<<
 *         elif size == 0x09:
 *             write(b"\xA9")
 */
  __pyx_tuple__124 = PyTuple_Pack(1, __pyx_kp_b__123); if (unlikely(!__pyx_tuple__124)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);

  /* "neo4j/packstream/_packer.pyx":265
 *             write(b"\xA8")
 *         elif size == 0x09:
 *             write(b"\xA9")             # <<<<<<<<<<<<<<
 *         elif size == 0x0A:
 *             write(b"\xAA")
 */
  __pyx_tuple__126 = PyTuple_Pack(1, __pyx_kp_b__125); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);

  /* "neo4j/packstream/_packer.pyx":267
 *             write(b"\xA9")
 *         elif size == 0x0A:
 *             write(b"\xAA")             # <<<<<<<<<<<<<<
 *         elif size == 0x0B:
 *             write(b"\xAB")
 */
  __pyx_tuple__128 = PyTuple_Pack(1, __pyx_kp_b__127); if (unlikely(!__pyx_tuple__128)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);

  /* "neo4j/packstream/_packer.pyx":269
 *             write(b"\xAA")
 *         elif size == 0x0B:
 *             write(b"\xAB")             # <<<<<<<<<<<<<<
 *         elif size == 0x0C:
 *             write(b"\xAC")
 */
  __pyx_tuple__130 = PyTuple_Pack(1, __pyx_kp_b__129); if (unlikely(!__pyx_tuple__130)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);

  /* "neo4j/packstream/_packer.pyx":271
 *             write(b"\xAB")
 *         elif size == 0x0C:
 *             write(b"\xAC")             # <<<<<<<<<<<<<<
 *         elif size == 0x0D:
 *             write(b"\xAD")
 */
  __pyx_tuple__132 = PyTuple_Pack(1, __pyx_kp_b__131); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);

  /* "neo4j/packstream/_packer.pyx":273
 *             write(b"\xAC")
 *         elif size == 0x0D:
 *             write(b"\xAD")             # <<<<<<<<<<<<<<
 *         elif size == 0x0E:
 *             write(b"\xAE")
 */
  __pyx_tuple__134 = PyTuple_Pack(1, __pyx_kp_b__133); if (unlikely(!__pyx_tuple__134)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);

  /* "neo4j/packstream/_packer.pyx":275
 *             write(b"\xAD")
 *         elif size == 0x0E:
 *             write(b"\xAE")             # <<<<<<<<<<<<<<
 *         elif size == 0x0F:
 *             write(b"\xAF")
 */
  __pyx_tuple__136 = PyTuple_Pack(1, __pyx_kp_b__135); if (unlikely(!__pyx_tuple__136)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__136);
  __Pyx_GIVEREF(__pyx_tuple__136);

  /* "neo4j/packstream/_packer.pyx":277
 *             write(b"\xAE")
 *         elif size == 0x0F:
 *             write(b"\xAF")             # <<<<<<<<<<<<<<
 *         elif size < 0x100:
 *             write(b"\xD8")
 */
  __pyx_tuple__138 = PyTuple_Pack(1, __pyx_kp_b__137); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);

  /* "neo4j/packstream/_packer.pyx":279
 *             write(b"\xAF")
 *         elif size < 0x100:
 *             write(b"\xD8")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 */
  __pyx_tuple__140 = PyTuple_Pack(1, __pyx_kp_b__139); if (unlikely(!__pyx_tuple__140)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);

  /* "neo4j/packstream/_packer.pyx":282
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 *             write(b"\xD9")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 */
  __pyx_tuple__142 = PyTuple_Pack(1, __pyx_kp_b__141); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);

  /* "neo4j/packstream/_packer.pyx":285
 *             write(PACKED_UINT_16[size])
 *         elif size < 0x100000000:
 *             write(b"\xDA")             # <<<<<<<<<<<<<<
 *             write(struct_pack(">I", size))
 *         else:
 */
  __pyx_tuple__144 = PyTuple_Pack(1, __pyx_kp_b__143); if (unlikely(!__pyx_tuple__144)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);

  /* "neo4j/packstream/_packer.pyx":288
 *             write(struct_pack(">I", size))
 *         else:
 *             raise OverflowError("Map header size out of range")             # <<<<<<<<<<<<<<
 * 
 *     cpdef pack_map_stream_header(self):
 */
  __pyx_tuple__145 = PyTuple_Pack(1, __pyx_kp_s_Map_header_size_out_of_range); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);

  /* "neo4j/packstream/_packer.pyx":291
 * 
 *     cpdef pack_map_stream_header(self):
 *         self._write(b"\xDB")             # <<<<<<<<<<<<<<
 * 
 *     cpdef pack_struct(self, bytes signature, tuple fields):
 */
  __pyx_tuple__147 = PyTuple_Pack(1, __pyx_kp_b__146); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);

  /* "neo4j/packstream/_packer.pyx":295
 *     cpdef pack_struct(self, bytes signature, tuple fields):
 *         if len(signature) != 1:
 *             raise ValueError("Structure signature must be a single byte value")             # <<<<<<<<<<<<<<
 *         write = self._write
 *         size = len(fields)
 */
  __pyx_tuple__148 = PyTuple_Pack(1, __pyx_kp_s_Structure_signature_must_be_a_si); if (unlikely(!__pyx_tuple__148)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);

  /* "neo4j/packstream/_packer.pyx":299
 *         size = len(fields)
 *         if size == 0x00:
 *             write(b"\xB0")             # <<<<<<<<<<<<<<
 *         elif size == 0x01:
 *             write(b"\xB1")
 */
  __pyx_tuple__150 = PyTuple_Pack(1, __pyx_kp_b__149); if (unlikely(!__pyx_tuple__150)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__150);
  __Pyx_GIVEREF(__pyx_tuple__150);

  /* "neo4j/packstream/_packer.pyx":301
 *             write(b"\xB0")
 *         elif size == 0x01:
 *             write(b"\xB1")             # <<<<<<<<<<<<<<
 *         elif size == 0x02:
 *             write(b"\xB2")
 */
  __pyx_tuple__152 = PyTuple_Pack(1, __pyx_kp_b__151); if (unlikely(!__pyx_tuple__152)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__152);
  __Pyx_GIVEREF(__pyx_tuple__152);

  /* "neo4j/packstream/_packer.pyx":303
 *             write(b"\xB1")
 *         elif size == 0x02:
 *             write(b"\xB2")             # <<<<<<<<<<<<<<
 *         elif size == 0x03:
 *             write(b"\xB3")
 */
  __pyx_tuple__154 = PyTuple_Pack(1, __pyx_kp_b__153); if (unlikely(!__pyx_tuple__154)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__154);
  __Pyx_GIVEREF(__pyx_tuple__154);

  /* "neo4j/packstream/_packer.pyx":305
 *             write(b"\xB2")
 *         elif size == 0x03:
 *             write(b"\xB3")             # <<<<<<<<<<<<<<
 *         elif size == 0x04:
 *             write(b"\xB4")
 */
  __pyx_tuple__156 = PyTuple_Pack(1, __pyx_kp_b__155); if (unlikely(!__pyx_tuple__156)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__156);
  __Pyx_GIVEREF(__pyx_tuple__156);

  /* "neo4j/packstream/_packer.pyx":307
 *             write(b"\xB3")
 *         elif size == 0x04:
 *             write(b"\xB4")             # <<<<<<<<<<<<<<
 *         elif size == 0x05:
 *             write(b"\xB5")
 */
  __pyx_tuple__158 = PyTuple_Pack(1, __pyx_kp_b__157); if (unlikely(!__pyx_tuple__158)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__158);
  __Pyx_GIVEREF(__pyx_tuple__158);

  /* "neo4j/packstream/_packer.pyx":309
 *             write(b"\xB4")
 *         elif size == 0x05:
 *             write(b"\xB5")             # <<<<<<<<<<<<<<
 *         elif size == 0x06:
 *             write(b"\xB6")
 */
  __pyx_tuple__160 = PyTuple_Pack(1, __pyx_kp_b__159); if (unlikely(!__pyx_tuple__160)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__160);
  __Pyx_GIVEREF(__pyx_tuple__160);

  /* "neo4j/packstream/_packer.pyx":311
 *             write(b"\xB5")
 *         elif size == 0x06:
 *             write(b"\xB6")             # <<<<<<<<<<<<<<
 *         elif size == 0x07:
 *             write(b"\xB7")
 */
  __pyx_tuple__162 = PyTuple_Pack(1, __pyx_kp_b__161); if (unlikely(!__pyx_tuple__162)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__162);
  __Pyx_GIVEREF(__pyx_tuple__162);

  /* "neo4j/packstream/_packer.pyx":313
 *             write(b"\xB6")
 *         elif size == 0x07:
 *             write(b"\xB7")             # <<<<<<<<<<<<<<
 *         elif size == 0x08:
 *             write(b"\xB8")
 */
  __pyx_tuple__164 = PyTuple_Pack(1, __pyx_kp_b__163); if (unlikely(!__pyx_tuple__164)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__164);
  __Pyx_GIVEREF(__pyx_tuple__164);

  /* "neo4j/packstream/_packer.pyx":315
 *             write(b"\xB7")
 *         elif size == 0x08:
 *             write(b"\xB8")             # <<<<<<<<<<<<<<
 *         elif size == 0x09:
 *             write(b"\xB9")
 */
  __pyx_tuple__166 = PyTuple_Pack(1, __pyx_kp_b__165); if (unlikely(!__pyx_tuple__166)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__166);
  __Pyx_GIVEREF(__pyx_tuple__166);

  /* "neo4j/packstream/_packer.pyx":317
 *             write(b"\xB8")
 *         elif size == 0x09:
 *             write(b"\xB9")             # <<<<<<<<<<<<<<
 *         elif size == 0x0A:
 *             write(b"\xBA")
 */
  __pyx_tuple__168 = PyTuple_Pack(1, __pyx_kp_b__167); if (unlikely(!__pyx_tuple__168)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__168);
  __Pyx_GIVEREF(__pyx_tuple__168);

  /* "neo4j/packstream/_packer.pyx":319
 *             write(b"\xB9")
 *         elif size == 0x0A:
 *             write(b"\xBA")             # <<<<<<<<<<<<<<
 *         elif size == 0x0B:
 *             write(b"\xBB")
 */
  __pyx_tuple__170 = PyTuple_Pack(1, __pyx_kp_b__169); if (unlikely(!__pyx_tuple__170)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__170);
  __Pyx_GIVEREF(__pyx_tuple__170);

  /* "neo4j/packstream/_packer.pyx":321
 *             write(b"\xBA")
 *         elif size == 0x0B:
 *             write(b"\xBB")             # <<<<<<<<<<<<<<
 *         elif size == 0x0C:
 *             write(b"\xBC")
 */
  __pyx_tuple__172 = PyTuple_Pack(1, __pyx_kp_b__171); if (unlikely(!__pyx_tuple__172)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__172);
  __Pyx_GIVEREF(__pyx_tuple__172);

  /* "neo4j/packstream/_packer.pyx":323
 *             write(b"\xBB")
 *         elif size == 0x0C:
 *             write(b"\xBC")             # <<<<<<<<<<<<<<
 *         elif size == 0x0D:
 *             write(b"\xBD")
 */
  __pyx_tuple__174 = PyTuple_Pack(1, __pyx_kp_b__173); if (unlikely(!__pyx_tuple__174)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__174);
  __Pyx_GIVEREF(__pyx_tuple__174);

  /* "neo4j/packstream/_packer.pyx":325
 *             write(b"\xBC")
 *         elif size == 0x0D:
 *             write(b"\xBD")             # <<<<<<<<<<<<<<
 *         elif size == 0x0E:
 *             write(b"\xBE")
 */
  __pyx_tuple__176 = PyTuple_Pack(1, __pyx_kp_b__175); if (unlikely(!__pyx_tuple__176)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__176);
  __Pyx_GIVEREF(__pyx_tuple__176);

  /* "neo4j/packstream/_packer.pyx":327
 *             write(b"\xBD")
 *         elif size == 0x0E:
 *             write(b"\xBE")             # <<<<<<<<<<<<<<
 *         elif size == 0x0F:
 *             write(b"\xBF")
 */
  __pyx_tuple__178 = PyTuple_Pack(1, __pyx_kp_b__177); if (unlikely(!__pyx_tuple__178)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__178);
  __Pyx_GIVEREF(__pyx_tuple__178);

  /* "neo4j/packstream/_packer.pyx":329
 *             write(b"\xBE")
 *         elif size == 0x0F:
 *             write(b"\xBF")             # <<<<<<<<<<<<<<
 *         elif size < 0x100:
 *             write(b"\xDC")
 */
  __pyx_tuple__180 = PyTuple_Pack(1, __pyx_kp_b__179); if (unlikely(!__pyx_tuple__180)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__180);
  __Pyx_GIVEREF(__pyx_tuple__180);

  /* "neo4j/packstream/_packer.pyx":331
 *             write(b"\xBF")
 *         elif size < 0x100:
 *             write(b"\xDC")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 */
  __pyx_tuple__182 = PyTuple_Pack(1, __pyx_kp_b__181); if (unlikely(!__pyx_tuple__182)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__182);
  __Pyx_GIVEREF(__pyx_tuple__182);

  /* "neo4j/packstream/_packer.pyx":334
 *             write(PACKED_UINT_8[size])
 *         elif size < 0x10000:
 *             write(b"\xDD")             # <<<<<<<<<<<<<<
 *             write(PACKED_UINT_16[size])
 *         else:
 */
  __pyx_tuple__184 = PyTuple_Pack(1, __pyx_kp_b__183); if (unlikely(!__pyx_tuple__184)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__184);
  __Pyx_GIVEREF(__pyx_tuple__184);

  /* "neo4j/packstream/_packer.pyx":337
 *             write(PACKED_UINT_16[size])
 *         else:
 *             raise OverflowError("Structure size out of range")             # <<<<<<<<<<<<<<
 *         write(signature)
 *         for field in fields:
 */
  __pyx_tuple__185 = PyTuple_Pack(1, __pyx_kp_s_Structure_size_out_of_range); if (unlikely(!__pyx_tuple__185)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__185);
  __Pyx_GIVEREF(__pyx_tuple__185);

  /* "neo4j/packstream/_packer.pyx":343
 * 
 *     cpdef pack_end_of_stream(self):
 *         self._write(b"\xDF")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__187 = PyTuple_Pack(1, __pyx_kp_b__186); if (unlikely(!__pyx_tuple__187)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__187);
  __Pyx_GIVEREF(__pyx_tuple__187);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_128 = PyInt_FromLong(128); if (unlikely(!__pyx_int_128)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_256 = PyInt_FromLong(256); if (unlikely(!__pyx_int_256)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_32768 = PyInt_FromLong(32768L); if (unlikely(!__pyx_int_32768)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_65536 = PyInt_FromLong(65536L); if (unlikely(!__pyx_int_65536)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2147483648 = PyInt_FromString((char *)"2147483648", 0, 0); if (unlikely(!__pyx_int_2147483648)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4294967296 = PyInt_FromString((char *)"4294967296", 0, 0); if (unlikely(!__pyx_int_4294967296)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_9223372036854775808 = PyInt_FromString((char *)"9223372036854775808", 0, 0); if (unlikely(!__pyx_int_9223372036854775808)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_16 = PyInt_FromLong(-16); if (unlikely(!__pyx_int_neg_16)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_128 = PyInt_FromLong(-128); if (unlikely(!__pyx_int_neg_128)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_32768 = PyInt_FromLong(-32768L); if (unlikely(!__pyx_int_neg_32768)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_2147483648 = PyInt_FromLong(-2147483648L); if (unlikely(!__pyx_int_neg_2147483648)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_9223372036854775808 = PyInt_FromString((char *)"-9223372036854775808", 0, 0); if (unlikely(!__pyx_int_neg_9223372036854775808)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_packer(void); /*proto*/
PyMODINIT_FUNC init_packer(void)
#else
PyMODINIT_FUNC PyInit__packer(void); /*proto*/
PyMODINIT_FUNC PyInit__packer(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit__packer(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_packer", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_neo4j__packstream___packer) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "neo4j.packstream._packer")) {
      if (unlikely(PyDict_SetItemString(modules, "neo4j.packstream._packer", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  __pyx_v_5neo4j_10packstream_7_packer_INT64_HI = Py_None; Py_INCREF(Py_None);
  __pyx_v_5neo4j_10packstream_7_packer_INT64_LO = Py_None; Py_INCREF(Py_None);
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  __pyx_vtabptr_5neo4j_10packstream_7_packer_Packer = &__pyx_vtable_5neo4j_10packstream_7_packer_Packer;
  __pyx_vtable_5neo4j_10packstream_7_packer_Packer.pack_raw = (PyObject *(*)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, PyObject *))__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_raw;
  __pyx_vtable_5neo4j_10packstream_7_packer_Packer.pack = (PyObject *(*)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack;
  __pyx_vtable_5neo4j_10packstream_7_packer_Packer._pack = (PyObject *(*)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, PyObject *))__pyx_f_5neo4j_10packstream_7_packer_6Packer__pack;
  __pyx_vtable_5neo4j_10packstream_7_packer_Packer.pack_bytes_header = (PyObject *(*)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int))__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_bytes_header;
  __pyx_vtable_5neo4j_10packstream_7_packer_Packer.pack_string_header = (PyObject *(*)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int))__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_string_header;
  __pyx_vtable_5neo4j_10packstream_7_packer_Packer.pack_list_header = (PyObject *(*)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int))__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_list_header;
  __pyx_vtable_5neo4j_10packstream_7_packer_Packer.pack_list_stream_header = (PyObject *(*)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int __pyx_skip_dispatch))__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_list_stream_header;
  __pyx_vtable_5neo4j_10packstream_7_packer_Packer.pack_map_header = (PyObject *(*)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int))__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_map_header;
  __pyx_vtable_5neo4j_10packstream_7_packer_Packer.pack_map_stream_header = (PyObject *(*)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int __pyx_skip_dispatch))__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_map_stream_header;
  __pyx_vtable_5neo4j_10packstream_7_packer_Packer.pack_struct = (PyObject *(*)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_struct;
  __pyx_vtable_5neo4j_10packstream_7_packer_Packer.pack_end_of_stream = (PyObject *(*)(struct __pyx_obj_5neo4j_10packstream_7_packer_Packer *, int __pyx_skip_dispatch))__pyx_f_5neo4j_10packstream_7_packer_6Packer_pack_end_of_stream;
  if (PyType_Ready(&__pyx_type_5neo4j_10packstream_7_packer_Packer) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __pyx_type_5neo4j_10packstream_7_packer_Packer.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_5neo4j_10packstream_7_packer_Packer.tp_dict, __pyx_vtabptr_5neo4j_10packstream_7_packer_Packer) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "Packer", (PyObject *)&__pyx_type_5neo4j_10packstream_7_packer_Packer) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __pyx_ptype_5neo4j_10packstream_7_packer_Packer = &__pyx_type_5neo4j_10packstream_7_packer_Packer;
  /*--- Type import code ---*/
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "neo4j/packstream/_packer.pyx":22
 * 
 * 
 * from struct import pack as struct_pack             # <<<<<<<<<<<<<<
 * 
 * from neo4j.compat import integer, string
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_pack);
  __Pyx_GIVEREF(__pyx_n_s_pack);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_pack);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_struct, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_struct_pack, __pyx_t_1) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "neo4j/packstream/_packer.pyx":24
 * from struct import pack as struct_pack
 * 
 * from neo4j.compat import integer, string             # <<<<<<<<<<<<<<
 * 
 * from .constants import *
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_integer);
  __Pyx_GIVEREF(__pyx_n_s_integer);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_integer);
  __Pyx_INCREF(__pyx_n_s_string);
  __Pyx_GIVEREF(__pyx_n_s_string);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_string);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_neo4j_compat, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_integer); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_integer, __pyx_t_2) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_string, __pyx_t_2) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":26
 * from neo4j.compat import integer, string
 * 
 * from .constants import *             # <<<<<<<<<<<<<<
 * from .structure import Structure
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s__188);
  __Pyx_GIVEREF(__pyx_n_s__188);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s__188);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_constants, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_import_star(__pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "neo4j/packstream/_packer.pyx":27
 * 
 * from .constants import *
 * from .structure import Structure             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Structure);
  __Pyx_GIVEREF(__pyx_n_s_Structure);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Structure);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_structure, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Structure); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Structure, __pyx_t_2) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "neo4j/packstream/_packer.pyx":30
 * 
 * 
 * cdef INT64_HI = 2 ** 63             # <<<<<<<<<<<<<<
 * cdef INT64_LO = -(2 ** 63)
 * 
 */
  __Pyx_INCREF(__pyx_int_9223372036854775808);
  __Pyx_XGOTREF(__pyx_v_5neo4j_10packstream_7_packer_INT64_HI);
  __Pyx_DECREF_SET(__pyx_v_5neo4j_10packstream_7_packer_INT64_HI, __pyx_int_9223372036854775808);
  __Pyx_GIVEREF(__pyx_int_9223372036854775808);

  /* "neo4j/packstream/_packer.pyx":31
 * 
 * cdef INT64_HI = 2 ** 63
 * cdef INT64_LO = -(2 ** 63)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_int_neg_9223372036854775808);
  __Pyx_XGOTREF(__pyx_v_5neo4j_10packstream_7_packer_INT64_LO);
  __Pyx_DECREF_SET(__pyx_v_5neo4j_10packstream_7_packer_INT64_LO, __pyx_int_neg_9223372036854775808);
  __Pyx_GIVEREF(__pyx_int_neg_9223372036854775808);

  /* "neo4j/packstream/_packer.pyx":1
 * #!/usr/bin/env python             # <<<<<<<<<<<<<<
 * # -*- encoding: utf-8 -*-
 * 
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init neo4j.packstream._packer", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init neo4j.packstream._packer");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs, NULL);
}
#endif  // CYTHON_FAST_PYCCALL

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = PyThreadState_GET();
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif  // CPython < 3.6
#endif  // CYTHON_FAST_PYCALL

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* GetModuleGlobalName */
  static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_RemainderObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = a % b;
            x += ((x != 0) & ((x ^ b) < 0)) * b;
            return PyInt_FromLong(x);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                default: return PyLong_Type.tp_as_number->nb_remainder(op1, op2);
            }
        }
                x = a % b;
                x += ((x != 0) & ((x ^ b) < 0)) * b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla % llb;
                llx += ((llx != 0) & ((llx ^ llb) < 0)) * llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceRemainder : PyNumber_Remainder)(op1, op2);
}
#endif

/* PyErrFetchRestore */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
    #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectCallNoArg */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* RaiseTooManyValuesToUnpack */
        static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
        static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
        static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
        static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* SaveResetException */
        #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
        #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    return PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
        #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* GetItemInt */
          static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ArgTypeTest */
          static void __Pyx_RaiseArgumentTypeInvalid(const char* name, PyObject *obj, PyTypeObject *type) {
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
}
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (likely(Py_TYPE(obj) == type)) return 1;
        #if PY_MAJOR_VERSION == 2
        else if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(PyObject_TypeCheck(obj, type))) return 1;
    }
    __Pyx_RaiseArgumentTypeInvalid(name, obj, type);
    return 0;
}

/* SetVTable */
          static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* Import */
          static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
          static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CodeObjectCache */
          static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
          #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
          #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
          static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CStringEquals */
          static CYTHON_INLINE int __Pyx_StrEq(const char *s1, const char *s2) {
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    return *s1 == *s2;
}

/* CheckBinaryVersion */
          static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
          static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
  #else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
  #endif
#else
  res = PyNumber_Int(x);
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
